/*	$OpenBSD: malo.c,v 1.71 2007/05/29 18:03:25 claudio Exp $ */

/*-
 * Copyright (c) 2006 Claudio Jeker <claudio@openbsd.org>
 * Copyright (c) 2006 Marcus Glocker <mglocker@openbsd.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/param.h>
#include <sys/systm.h> 
#include <sys/module.h>
#include <sys/kernel.h>
#include <sys/lock.h>
#include <sys/linker.h>
#include <sys/firmware.h>
#include <sys/mutex.h>
#include <sys/taskqueue.h>
#include <sys/endian.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <sys/callout.h>

#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/bus.h>
#include <sys/rman.h>

#include <sys/socket.h>
 
#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_media.h>
#include <net/if_types.h>
#include <net/if_arp.h>
#include <net/ethernet.h>

#include <net80211/ieee80211_var.h>
#include <net80211/ieee80211_radiotap.h>

#include <net/bpf.h>

#include <dev/pci/pcivar.h>
#include <dev/pci/pcireg.h>

#if 0
#include <dev/malo/if_malo.h>
#else
#include "if_malo.h"
#endif

#define	MALO_DEBUG

#ifdef MALO_DEBUG
#define DPRINTF(sc, ...)	do {				\
	if (malo_debug > 0)					\
		device_printf((sc)->sc_dev, __VA_ARGS__);	\
} while (0)
#define DPRINTFN(sc, n, ...)	do {				\
	if (malo_debug >= (n))					\
		device_printf((sc)->sc_dev, __VA_ARGS__);	\
} while (0)
/*int malo_debug = 1;*/
int malo_debug = 1;
#else
#define DPRINTF(sc, ...)
#define DPRINTFN(sc, n, ...)
#endif

/* Tunables. */
static int msi_disable = 0;
TUNABLE_INT("hw.malo.msi_disble", &msi_disable);
static int legacy_intr = 0;
TUNABLE_INT("hw.malo.legacy_intr", &legacy_intr);

/*
 * Devices supported by this driver.
 */
static struct malo_product {
	uint16_t	malo_vendorid;
	uint16_t	malo_deviceid;
	const char	*malo_name;
} malo_products[] = {
	{ VENDORID_MARVELL, DEVICEID_MRVL_88W8310,
	    "Marvell Libertas 88W8310 802.11g Wireless Adapter" },
	{ VENDORID_MARVELL, DEVICEID_MRVL_88W8335R1,
	    "Marvell Libertas 88W8335 802.11g Wireless Adapter" },
	{ VENDORID_MARVELL, DEVICEID_MRVL_88W8335R2,
	    "Marvell Libertas 88W8335 802.11g Wireless Adapter" }
};

static struct resource_spec malo_res_spec_mem[] = {
	{ SYS_RES_MEMORY,	PCIR_BAR(0),	RF_ACTIVE },
	{ SYS_RES_MEMORY,	PCIR_BAR(1),	RF_ACTIVE },
	{ -1,			0,		0 }
};

static struct resource_spec malo_res_spec_legacy[] = {
	{ SYS_RES_IRQ,		0,		RF_ACTIVE | RF_SHAREABLE },
	{ -1,			0,		0 }
};

static struct resource_spec malo_res_spec_msi[] = {
	{ SYS_RES_IRQ,		1,		RF_ACTIVE },
	{ -1,			0,		0 }
};


static int	malo_probe(device_t);
static int	malo_attach(device_t);
static int	malo_detach(device_t);
static int	malo_dma_alloc(struct malo_softc *);
static void	malo_dma_free(struct malo_softc *);
static void	malo_scan_start(struct ieee80211com *);
static void	malo_scan_end(struct ieee80211com *);
static void	malo_set_channel(struct ieee80211com *);

static void	malo_legacy_intr(void *);
static __inline void	malo_enable_intr(struct malo_softc *);
static __inline void	malo_disable_intr(struct malo_softc *);
static __inline void	malo_discard_rxbuf(struct malo_softc *, int);
static int	malo_newbuf(struct malo_softc *, int);
static int	malo_init_tx_ring(struct malo_softc *);
static int	malo_init_rx_ring(struct malo_softc *);

static void	malo_send_cmd(struct malo_softc *, bus_addr_t);
static int	malo_send_cmd_dma(struct malo_softc *, bus_addr_t);
static void	malo_init(void *);
static void	malo_init_locked(struct malo_softc *);
static int	malo_ioctl(struct ifnet *, u_long, caddr_t);
static void	malo_start(struct ifnet *);
static void	malo_start_locked(struct ifnet *);
static void	malo_stop(struct malo_softc *);
static void	malo_watchdog(struct ifnet *);
static int	malo_newstate(struct ieee80211com *, enum ieee80211_state, int);
static void	malo_newassoc(struct ieee80211_node *, int);
static struct ieee80211_node *
	malo_node_alloc(struct ieee80211_node_table *);
static int	malo_media_change(struct ifnet *);
static void	malo_media_status(struct ifnet *, struct ifmediareq *);
static int	malo_chip2rate(int);
static int	malo_fix2rate(int);
#if 0
static void	malo_next_scan(void *);
#endif
static void	malo_tx_intr(struct malo_softc *);
static int	malo_tx_mgt(struct malo_softc *, struct mbuf **,
	    struct ieee80211_node *);
static int	malo_tx_data(struct malo_softc *, struct mbuf **,
	    struct ieee80211_node *);
static void	malo_rx_intr(struct malo_softc *);
static int	malo_load_bootimg(struct malo_softc *);
static int	malo_load_firmware(struct malo_softc *);

static int	malo_set_wepkey(struct malo_softc *);
static int	malo_set_slot(struct malo_softc *);
static void	malo_update_slot(struct ifnet *);
#ifdef MALO_DEBUG
static void	malo_hexdump(void *, int);
#endif
static char *
	malo_cmd_string(uint16_t cmd);
static char *
	malo_cmd_string_result(uint16_t result);
static int	malo_cmd_get_spec(struct malo_softc *);
static int	malo_cmd_set_wepkey(struct malo_softc *, struct ieee80211_key *,
	    uint16_t);
static int	malo_cmd_set_prescan(struct malo_softc *);
static int	malo_cmd_set_postscan(struct malo_softc *, uint8_t *, uint8_t);
static int	malo_cmd_set_channel(struct malo_softc *, uint8_t);
static int	malo_cmd_set_antenna(struct malo_softc *, uint16_t);
static int	malo_cmd_set_radio(struct malo_softc *, uint16_t, uint16_t);
static int	malo_cmd_set_aid(struct malo_softc *, uint8_t *, uint16_t);
static int	malo_cmd_set_txpower(struct malo_softc *, unsigned int);
static int	malo_cmd_set_rts(struct malo_softc *, uint32_t);
static int	malo_cmd_set_slot(struct malo_softc *, uint8_t);
static int	malo_cmd_set_rate(struct malo_softc *, uint8_t);

/*static struct mbuf * malo_fragment(struct mbuf *, int, int);*/

static int
malo_intr(void *xsc)
{
	printf("INTR ");
	return (FILTER_HANDLED);
}

static int
malo_suspend(struct malo_softc *sc)
{
	return (0);
}

static int
malo_resume(struct malo_softc *sc)
{
	return (0);
}

static int
malo_shutdown(struct malo_softc *sc)
{
	return (0);
}

static int
malo_probe(device_t dev)
{
	struct malo_product *mp;
	uint16_t vendor, devid;
	int i;

	vendor = pci_get_vendor(dev);
	devid = pci_get_device(dev);
	mp = malo_products;
	for (i = 0; i < sizeof(malo_products)/sizeof(malo_products[0]);
	    i++, mp++) {
		if (vendor == mp->malo_vendorid && devid == mp->malo_deviceid) {
			device_set_desc(dev, mp->malo_name);
			return (BUS_PROBE_DEFAULT);
		}
	}

	return (ENXIO);
}

static int
malo_attach(device_t dev)
{
	struct malo_softc *sc;
	struct ieee80211com *ic;
	struct ieee80211_channel *c;
	struct ifnet *ifp;
	int error, i, msic;

	sc = device_get_softc(dev);
	sc->sc_dev = dev;

	mtx_init(&sc->sc_mtx, device_get_nameunit(dev), MTX_NETWORK_LOCK,
	    MTX_DEF);
	callout_init_mtx(&sc->sc_scan_to, &sc->sc_mtx, 0);

	/*
	 * Map control/status registers.
	 */
	pci_enable_busmaster(dev);

	/* Allocate memory resource */
	sc->sc_mem_spec = malo_res_spec_mem;
	error = bus_alloc_resources(dev, sc->sc_mem_spec, sc->sc_res_mem);
	if (error) {
		device_printf(dev, "couldn't allocate memory resources\n");
		mtx_destroy(&sc->sc_mtx);
		return (ENXIO);
	}

	/* Allocate IRQ resources. */
	sc->sc_irq_spec = malo_res_spec_legacy;
	msic = pci_msi_count(dev);
	if (bootverbose)
		device_printf(dev, "MSI count : %d\n", msic);
	if (legacy_intr != 0)
		msi_disable = 1;
	if (msic == 1 && msi_disable == 0 && pci_alloc_msi(dev, &msic) == 0) {
		if (msic == 2) {
			sc->sc_msi = 1;
			sc->sc_irq_spec = malo_res_spec_msi;
		} else
			pci_release_msi(dev);
	}

	error = bus_alloc_resources(dev, sc->sc_irq_spec, sc->sc_res_irq);
	if (error) {
		device_printf(dev, "couldn't allocate IRQ resources\n");
		goto fail;
	}

	/* allocate DMA structures */
	malo_dma_alloc(sc);

#if 0
	MALO_LOCK(sc);
	if ((error = malo_load_bootimg(sc)) ||
	    (error = malo_load_firmware(sc))) {
		MALO_UNLOCK(sc);
		goto fail;
	}
#endif

	ifp = sc->sc_ifp = if_alloc(IFT_ETHER);
	if (ifp == NULL) {
		device_printf(dev, "couldn't if_alloc()\n");
		error = ENOSPC;
		goto fail;
	}

	/* setup interface */
	if_initname(ifp, device_get_name(sc->sc_dev),
	    device_get_unit(sc->sc_dev));
	ifp->if_softc = sc;
	ifp->if_init = malo_init;
	ifp->if_ioctl = malo_ioctl;
	ifp->if_start = malo_start;
	ifp->if_watchdog = malo_watchdog;
	ifp->if_flags = IFF_SIMPLEX | IFF_BROADCAST | IFF_MULTICAST;
	IFQ_SET_MAXLEN(&ifp->if_snd, IFQ_MAXLEN);
	ifp->if_snd.ifq_drv_maxlen = IFQ_MAXLEN;
	IFQ_SET_READY(&ifp->if_snd);

	ic = &sc->sc_ic;
	ic->ic_ifp = ifp;
	/* XXX ic->ic_reset = malo_reset; */
	ic->ic_newassoc = malo_newassoc;
	ic->ic_updateslot = malo_update_slot;

	ic->ic_scan_start = malo_scan_start;
	ic->ic_scan_end = malo_scan_end;
	ic->ic_set_channel = malo_set_channel;

	/* XXX not right but it's not used anyware important */
	ic->ic_phytype = IEEE80211_T_OFDM;
	ic->ic_opmode = IEEE80211_M_STA;
	ic->ic_caps =
	    IEEE80211_C_IBSS |
	    IEEE80211_C_MONITOR |
	    IEEE80211_C_SHPREAMBLE |
	    IEEE80211_C_SHSLOT |
	    IEEE80211_C_WEP;

	/* set supported rates */
	/* XXX
	ic->ic_sup_rates[IEEE80211_MODE_11B] = ieee80211_std_rateset_11b;
	ic->ic_sup_rates[IEEE80211_MODE_11G] = ieee80211_std_rateset_11g;
	sc->sc_last_txrate = -1;
	*/

	/* set channels */
	for (i = 1; i <= 14; i++) {
		c = &ic->ic_channels[ic->ic_nchans++];
		c->ic_freq = ieee80211_ieee2mhz(i, IEEE80211_CHAN_2GHZ);
		c->ic_flags =
		    IEEE80211_CHAN_PUREG |
		    IEEE80211_CHAN_B |
		    IEEE80211_CHAN_G;
		c->ic_ieee = i;
		/* XXX ic_maxregpower, ic_maxpower, ic_minpower */
	}

	/* XXX malo_init : ic->ic_state = IEEE80211_S_INIT; */
	/* XXX ic->ic_max_rssi = 75; */
	for (i = 0; i < ETHER_ADDR_LEN; i++)
		ic->ic_myaddr[i] = malo_ctl_read1(sc, 0xa528 + i);

	/* attach interface */
	ieee80211_ifattach(ic);

	/* override default methods */
	sc->sc_newstate = ic->ic_newstate;
	ic->ic_newstate = malo_newstate;
	ic->ic_node_alloc = malo_node_alloc;
	/* XXX ic_scan_start, ic_scan_end, ic_set_channel */

	ieee80211_media_init(ic, malo_media_change, malo_media_status);

	bpfattach2(ifp, DLT_IEEE802_11_RADIO, 
	    sizeof(struct ieee80211_frame) + 64, &sc->sc_drvbpf);

	sc->sc_rxtap_len = sizeof(sc->sc_rxtapu);
	sc->sc_rxtap.wr_ihdr.it_len = htole16(sc->sc_rxtap_len);
	sc->sc_rxtap.wr_ihdr.it_present = htole32(MALO_RX_RADIOTAP_PRESENT);

	sc->sc_txtap_len = sizeof(sc->sc_txtapu);
	sc->sc_txtap.wt_ihdr.it_len = htole16(sc->sc_txtap_len);
	sc->sc_txtap.wt_ihdr.it_present = htole32(MALO_TX_RADIOTAP_PRESENT);

	if (legacy_intr == 0) {
		/*
		TASK_INIT(&sc->sc_int_task, 0, malo_int_task, sc);
		sc->sc_tq = taskqueue_create_fast("malo_taskq", M_WAITOK,
		    taskqueue_thread_enqueue, &sc->sc_tq);
		taskqueue_start_threads(&sc->sc_tq, 1, PI_NET, "%s taskq",
		    device_get_nameunit(sc->sc_dev));
		*/
	}
	/* Hook interrupt last */
	if (sc->sc_msi == 0)
		error = bus_setup_intr(dev, sc->sc_res_irq[0],
		    INTR_TYPE_NET | INTR_MPSAFE, NULL, malo_legacy_intr, sc,
		    &sc->sc_intrhand[0]);
	else {
		for (i = 0; i < MALO_MSI_MESSAGES; i++) {
			error = bus_setup_intr(dev, sc->sc_res_irq[i],
			    INTR_TYPE_NET | INTR_MPSAFE, malo_intr, NULL, sc,
			    &sc->sc_intrhand[i]);
			if (error != 0)
				break;
		}
	}

	if (error != 0) {
		device_printf(dev, "couldn't set up interrupt handler\n");
		/*
		if (legacy_intr == 0)
			taskqueue_free(sc->sc_tq);
		*/
		sc->sc_tq = NULL;
		goto fail;
	}
fail:
	if (error != 0)
		malo_detach(dev);
	else {
		if (bootverbose)
			ieee80211_announce(ic);
	}

	return (error);
}

static int
malo_detach(device_t dev)
{
	struct malo_softc *sc;
	struct ieee80211com *ic;
	struct ifnet *ifp;
	int i;

	sc = device_get_softc(dev);
	ic = &sc->sc_ic;
	ifp = sc->sc_ifp;

	if (device_is_attached(dev)) {
		MALO_LOCK(sc);
		/* XXX IFF_UP clear? */
		malo_stop(sc);
		MALO_UNLOCK(sc);
		callout_drain(&sc->sc_scan_to);
		/* XXX taskqueue drain goes here */
	}

	if (ifp != NULL) {
		ieee80211_ifdetach(ic);
		if_detach(ifp);
	}

	malo_dma_free(sc);

	if (sc->sc_tq != NULL)
		taskqueue_free(sc->sc_tq);
	for (i = 0; i < MALO_MSI_MESSAGES; i++) {
		if (sc->sc_intrhand[i] != NULL) {
			bus_teardown_intr(dev, sc->sc_res_irq[i],
			    sc->sc_intrhand[i]);
			sc->sc_intrhand[i] = NULL;
		}
	}
	bus_release_resources(dev, sc->sc_irq_spec, sc->sc_res_irq);
	if (sc->sc_msi)
		pci_release_msi(dev);
	bus_release_resources(dev, sc->sc_mem_spec, sc->sc_res_mem);
	mtx_destroy(&sc->sc_mtx);

	return (0);
}

static int
malo_pci_shutdown(device_t dev)
{
	struct malo_softc *sc;

	sc = device_get_softc(dev);
	malo_shutdown(sc);

	return (0);
}

static int
malo_pci_suspend(device_t dev)
{
	struct malo_softc *sc;

	sc = device_get_softc(dev);
	malo_suspend(sc);

	return (0);
}

static int
malo_pci_resume(device_t dev)
{
	struct malo_softc *sc;

	sc = device_get_softc(dev);
	malo_resume(sc);

	return (0);
}

static void
malo_legacy_intr(void *arg)
{
	struct malo_softc *sc;
	struct malo_cmdheader *hdr;
	uint32_t status;
	struct ifnet *ifp;

	sc = arg;
	MALO_LOCK(sc);

	status = malo_ctl_read4(sc, 0x0c30);
	if (status == 0xffffffff || status == 0)
		goto done;

	malo_disable_intr(sc);

	hdr = sc->sc_cmd_mem;
	ifp = sc->sc_ifp;

	if ((ifp->if_drv_flags & IFF_DRV_RUNNING) == 0) {
		/* XXX */
		device_printf(sc->sc_dev, "INTR while not in running state!\n");
		/* check for firmware command */
		if (status & 0x4) {
			if (le16toh(hdr->result) != MALO_CMD_RESULT_OK) {
				device_printf(sc->sc_dev,
				    "firmware cmd %s failed with %s\n",
				    malo_cmd_string(hdr->cmd),
				    malo_cmd_string_result(hdr->result));
			}
#ifdef MALO_DEBUG
			device_printf(sc->sc_dev, "cmd answer for %s=%s\n",
			    malo_cmd_string(hdr->cmd),
			    malo_cmd_string_result(hdr->result));
			if (malo_debug > 2)
				malo_hexdump(hdr, le16toh(hdr->size));
#endif
		}
		if (status & ~0x7)
			DPRINTF(sc, "unkown interrupt %x\n", status);
		goto done;
	}

	for (; status != 0;) {
		if (status & 0x1)
			malo_tx_intr(sc);
		if (status & 0x2)
			malo_rx_intr(sc);
		if (status & 0x4) {
			if (le16toh(hdr->result) != MALO_CMD_RESULT_OK) {
				device_printf(sc->sc_dev,
				    "firmware cmd %s failed with %s\n",
				    malo_cmd_string(hdr->cmd),
				    malo_cmd_string_result(hdr->result));
			}
#ifdef MALO_DEBUG
			device_printf(sc->sc_dev, "cmd answer for %s=%s\n",
			    malo_cmd_string(hdr->cmd),
			    malo_cmd_string_result(hdr->result));
			if (malo_debug > 2)
				malo_hexdump(hdr, le16toh(hdr->size));
#endif
		}

		if (status & ~0x7)
			DPRINTF(sc, "unkown interrupt %x\n", status);
		status = malo_ctl_read4(sc, 0x0c30);
	}

	/* just ack the interrupt */
	/* XXX malo_ctl_write4(sc, 0x0c30, 0); */

done:
	malo_enable_intr(sc);
	MALO_UNLOCK(sc);
}

struct malo_dmamap_arg {
	bus_addr_t	malo_busaddr;
};

static void
malo_dmamap_cb(void *arg, bus_dma_segment_t *segs, int nseg, int error)
{
	struct malo_dmamap_arg *ctx;

	if (error != 0)
		return;

	ctx = (struct malo_dmamap_arg *)arg;
	ctx->malo_busaddr = segs[0].ds_addr;
}

static int
malo_dma_alloc(struct malo_softc *sc)
{
	struct malo_dmamap_arg ctx;
	struct malo_tx_data *txd;
	struct malo_rx_data *rxd;
	int error, i;

	/* create parent tag. */
	error = bus_dma_tag_create(bus_get_dma_tag(sc->sc_dev),/* parent */
	    1, 0,			/* algnmnt, boundary */
	    BUS_SPACE_MAXADDR_32BIT,	/* lowaddr */
	    BUS_SPACE_MAXADDR,		/* highaddr */
	    NULL, NULL,			/* filter, filterarg */
	    BUS_SPACE_MAXSIZE_32BIT,	/* maxsize */
	    0,				/* nsegments */
	    BUS_SPACE_MAXSIZE_32BIT,	/* maxsegsize */
	    0,				/* flags */
	    NULL, NULL,			/* lockfunc, lockarg */
	    &sc->sc_parent_tag);
	if (error != 0) {
		device_printf(sc->sc_dev, "failed to create parent DMA tag\n");
		goto fail;
	}

	/* create command block DMA tag */
	error = bus_dma_tag_create(sc->sc_parent_tag,	/* parent */
	    MALO_CMD_DMA_ALIGN, 0,	/* algnmnt, boundary */
	    BUS_SPACE_MAXADDR,		/* lowaddr */
	    BUS_SPACE_MAXADDR,		/* highaddr */
	    NULL, NULL,			/* filter, filterarg */
	    MALO_CMD_DMA_SZ,		/* maxsize */
	    1,				/* nsegments */
	    MALO_CMD_DMA_SZ,		/* maxsegsize */
	    0,				/* flags */
	    NULL, NULL,			/* lockfunc, lockarg */
	    &sc->sc_cmd_tag);
	if (error != 0) {
		device_printf(sc->sc_dev,
		    "failed to allocate command block DMA tag\n");
		goto fail;
	}

	/* create tag for Tx ring. */
	error = bus_dma_tag_create(sc->sc_parent_tag,	/* parent */
	    MALO_RING_ALIGN, 0,		/* algnmnt, boundary */
	    BUS_SPACE_MAXADDR,		/* lowaddr */
	    BUS_SPACE_MAXADDR,		/* highaddr */
	    NULL, NULL,			/* filter, filterarg */
	    MALO_TX_RING_SZ,		/* maxsize */
	    1,				/* nsegments */
	    MALO_TX_RING_SZ,		/* maxsegsize */
	    0,				/* flags */
	    NULL, NULL,			/* lockfunc, lockarg */
	    &sc->sc_txring.ring_tag);
	if (error != 0) {
		device_printf(sc->sc_dev,
		    "failed to allocate Tx ring DMA tag\n");
		goto fail;
	}

	/* create tag for Rx ring. */
	error = bus_dma_tag_create(sc->sc_parent_tag,	/* parent */
	    MALO_RING_ALIGN, 0,		/* algnmnt, boundary */
	    BUS_SPACE_MAXADDR,		/* lowaddr */
	    BUS_SPACE_MAXADDR,		/* highaddr */
	    NULL, NULL,			/* filter, filterarg */
	    MALO_RX_RING_SZ,		/* maxsize */
	    1,				/* nsegments */
	    MALO_RX_RING_SZ,		/* maxsegsize */
	    0,				/* flags */
	    NULL, NULL,			/* lockfunc, lockarg */
	    &sc->sc_rxring.ring_tag);
	if (error != 0) {
		device_printf(sc->sc_dev,
		    "failed to allocate Rx ring DMA tag\n");
		goto fail;
	}

	/* create tag for Tx buffers. */
	error = bus_dma_tag_create(sc->sc_parent_tag,	/* parent */
	    1, 0,			/* algnmnt, boundary */
	    BUS_SPACE_MAXADDR,		/* lowaddr */
	    BUS_SPACE_MAXADDR,		/* highaddr */
	    NULL, NULL,			/* filter, filterarg */
	    MALO_MAX_TX_SEGSIZE,	/* maxsize */
	    MALO_MAX_SCATTER,		/* nsegments */
	    MCLBYTES,			/* maxsegsize */
	    0,				/* flags */
	    NULL, NULL,			/* lockfunc, lockarg */
	    &sc->sc_txring.tx_tag);
	if (error != 0) {
		device_printf(sc->sc_dev, "failed to allocate Tx DMA tag\n");
		goto fail;
	}

	/* create tag for Rx buffers. */
	error = bus_dma_tag_create(sc->sc_parent_tag,	/* parent */
	    1, 0,			/* algnmnt, boundary */
	    BUS_SPACE_MAXADDR,		/* lowaddr */
	    BUS_SPACE_MAXADDR,		/* highaddr */
	    NULL, NULL,			/* filter, filterarg */
	    MCLBYTES,			/* maxsize */
	    1,				/* nsegments */
	    MCLBYTES,			/* maxsegsize */
	    0,				/* flags */
	    NULL, NULL,			/* lockfunc, lockarg */
	    &sc->sc_rxring.rx_tag);
	if (error != 0) {
		device_printf(sc->sc_dev, "failed to allocate Rx DMA tag\n");
		goto fail;
	}

	/* allocate DMA'able memory and load the DMA map for command block. */
	error = bus_dmamem_alloc(sc->sc_cmd_tag, (void **)&sc->sc_cmd_mem,
	    BUS_DMA_WAITOK | BUS_DMA_ZERO | BUS_DMA_COHERENT,
	    &sc->sc_cmd_map);
	if (error != 0) {
		device_printf(sc->sc_dev,
		    "failed to allocate DMA'able memory for command block\n");
		goto fail;
	}

	ctx.malo_busaddr = 0;
	error = bus_dmamap_load(sc->sc_cmd_tag, sc->sc_cmd_map,
	    sc->sc_cmd_mem, MALO_CMD_DMA_SZ, malo_dmamap_cb, &ctx, 0);
	if (error != 0 || ctx.malo_busaddr == 0) {
		device_printf(sc->sc_dev,
		    "failed to load DMA'able memory for command block\n");
		goto fail;
	}
	sc->sc_cmd_dmaaddr = ctx.malo_busaddr;
	/* initialize command DMA block */
	sc->sc_cookie = sc->sc_cmd_mem;
	*sc->sc_cookie = htole32(0xaa55aa55);
	sc->sc_cmd_mem = (char *)sc->sc_cmd_mem + sizeof(uint32_t);
	sc->sc_cookie_dmaaddr = sc->sc_cmd_dmaaddr;
	sc->sc_cmd_dmaaddr += sizeof(uint32_t);

	/* allocate DMA'able memory and load the DMA map for Tx ring. */
	error = bus_dmamem_alloc(sc->sc_txring.ring_tag,
	    (void **)&sc->sc_txring.desc,
	    BUS_DMA_WAITOK | BUS_DMA_ZERO | BUS_DMA_COHERENT,
	    &sc->sc_txring.ring_map);
	if (error != 0) {
		device_printf(sc->sc_dev,
		    "failed to allocate DMA'able memory for Tx ring\n");
		goto fail;
	}

	ctx.malo_busaddr = 0;
	error = bus_dmamap_load(sc->sc_txring.ring_tag, sc->sc_txring.ring_map,
	    sc->sc_txring.desc, MALO_TX_RING_SZ, malo_dmamap_cb, &ctx, 0);
	if (error != 0 || ctx.malo_busaddr == 0) {
		device_printf(sc->sc_dev,
		    "failed to load DMA'able memory for Tx ring\n");
		goto fail;
	}
	sc->sc_txring.physaddr = ctx.malo_busaddr;

	/* allocate DMA'able memory and load the DMA map for Rx ring. */
	error = bus_dmamem_alloc(sc->sc_rxring.ring_tag,
	    (void **)&sc->sc_rxring.desc,
	    BUS_DMA_WAITOK | BUS_DMA_ZERO | BUS_DMA_COHERENT,
	    &sc->sc_rxring.ring_map);
	if (error != 0) {
		device_printf(sc->sc_dev,
		    "failed to allocate DMA'able memory for Rx ring\n");
		goto fail;
	}

	ctx.malo_busaddr = 0;
	error = bus_dmamap_load(sc->sc_rxring.ring_tag, sc->sc_rxring.ring_map,
	    sc->sc_rxring.desc, MALO_RX_RING_SZ, malo_dmamap_cb, &ctx, 0);
	if (error != 0 || ctx.malo_busaddr == 0) {
		device_printf(sc->sc_dev,
		    "failed to load DMA'able memory for Rx ring\n");
		goto fail;
	}
	sc->sc_rxring.physaddr = ctx.malo_busaddr;

	/* create DMA maps for Tx buffers. */
	for (i = 0; i < MALO_TX_RING_COUNT; i++) {
		txd = &sc->sc_txring.data[i];
		txd->map = NULL;
		txd->m = NULL;
		txd->softstat = 0;
		txd->ni = NULL;
		error = bus_dmamap_create(sc->sc_txring.tx_tag, 0, &txd->map);
		if (error != 0) {
			device_printf(sc->sc_dev,
			    "failed to create Tx dmamap\n");
			goto fail;
		}
	}
	/* create DMA maps for Rx buffers. */
	if ((error = bus_dmamap_create(sc->sc_rxring.rx_tag, 0,
	    &sc->sc_rxring.rx_sparemap)) != 0) {
		device_printf(sc->sc_dev, "failed to create spare Rx dmamap\n");
		goto fail;
	}
	for (i = 0; i < MALO_RX_RING_COUNT; i++) {
		rxd = &sc->sc_rxring.data[i];
		rxd->map = NULL;
		rxd->m = NULL;
		error = bus_dmamap_create(sc->sc_rxring.rx_tag, 0, &rxd->map);
		if (error != 0) {
			device_printf(sc->sc_dev,
			    "failed to create Rx dmamap\n");
			goto fail;
		}
	}

fail:
	return (error);
}

static void
malo_dma_free(struct malo_softc *sc)
{
	struct malo_tx_data *txd;
	struct malo_rx_data *rxd;
	int i;

	/* Tx ring */
	if (sc->sc_txring.ring_tag != NULL) {
		if (sc->sc_txring.ring_map != NULL)
			bus_dmamap_unload(sc->sc_txring.ring_tag,
			    sc->sc_txring.ring_map);
		if (sc->sc_txring.ring_map != NULL &&
		    sc->sc_txring.desc != NULL)
			bus_dmamem_free(sc->sc_txring.ring_tag,
			    sc->sc_txring.desc, sc->sc_txring.ring_map);
		sc->sc_txring.desc = NULL;
		sc->sc_txring.ring_map = NULL;
		bus_dma_tag_destroy(sc->sc_txring.ring_tag);
		sc->sc_txring.ring_tag = NULL;
	}
	/* Rx ring */
	if (sc->sc_rxring.ring_tag != NULL) {
		if (sc->sc_rxring.ring_map != NULL)
			bus_dmamap_unload(sc->sc_rxring.ring_tag,
			    sc->sc_rxring.ring_map);
		if (sc->sc_rxring.ring_map != NULL &&
		    sc->sc_rxring.desc != NULL)
			bus_dmamem_free(sc->sc_rxring.ring_tag,
			    sc->sc_rxring.desc, sc->sc_rxring.ring_map);
		sc->sc_rxring.desc = NULL;
		sc->sc_rxring.ring_map = NULL;
		bus_dma_tag_destroy(sc->sc_rxring.ring_tag);
		sc->sc_rxring.ring_tag = NULL;
	}
	/* Tx buffers */
	if (sc->sc_txring.tx_tag != NULL) {
		for (i = 0; i < MALO_TX_RING_COUNT; i++) {
			txd = &sc->sc_txring.data[i];
			if (txd->map != NULL) {
				bus_dmamap_destroy(sc->sc_txring.tx_tag,
				    txd->map);
				txd->map = NULL;
			}
		}
		bus_dma_tag_destroy(sc->sc_txring.tx_tag);
		sc->sc_txring.tx_tag = NULL;
	}
	/* Rx buffers */
	if (sc->sc_rxring.rx_tag != NULL) {
		for (i = 0; i < MALO_RX_RING_COUNT; i++) {
			rxd = &sc->sc_rxring.data[i];
			if (rxd->map != NULL) {
				bus_dmamap_destroy(sc->sc_rxring.rx_tag,
				    rxd->map);
				rxd->map = NULL;
			}
		}
		if (sc->sc_rxring.rx_sparemap != NULL) {
			bus_dmamap_destroy(sc->sc_rxring.rx_tag,
			    sc->sc_rxring.rx_sparemap);
			sc->sc_rxring.rx_sparemap = NULL;
		}
		bus_dma_tag_destroy(sc->sc_rxring.rx_tag);
		sc->sc_rxring.rx_tag = NULL;
	}

	/* command DMA block */
	if (sc->sc_cmd_tag != NULL) {
		if (sc->sc_cmd_map != NULL) {
			/* XXX not sure dmamap sync is required here */
			bus_dmamap_sync(sc->sc_cmd_tag, sc->sc_cmd_map,
			    BUS_DMASYNC_POSTWRITE);
			bus_dmamap_unload(sc->sc_cmd_tag, sc->sc_cmd_map);
		}
		if (sc->sc_cmd_map != NULL &&
		    sc->sc_cookie != NULL)
			bus_dmamem_free(sc->sc_cmd_tag, sc->sc_cookie,
			    sc->sc_cmd_map);
		sc->sc_cmd_map = NULL;
		sc->sc_cmd_mem = NULL;
		sc->sc_cookie = NULL;
		bus_dma_tag_destroy(sc->sc_cmd_tag);
		sc->sc_cmd_tag = NULL;
	}

	if (sc->sc_parent_tag != NULL) {
		bus_dma_tag_destroy(sc->sc_parent_tag);
		sc->sc_parent_tag = NULL;
	}
}

static void
malo_send_cmd(struct malo_softc *sc, bus_addr_t addr)
{

	malo_ctl_write4(sc, 0x0c10, (uint32_t)addr);
	malo_ctl_barrier(sc, BUS_SPACE_BARRIER_WRITE);
	malo_ctl_write4(sc, 0x0c18, 2); /* CPU_TRANSFER_CMD */
	malo_ctl_barrier(sc, BUS_SPACE_BARRIER_WRITE);
}

static int
malo_send_cmd_dma(struct malo_softc *sc, bus_addr_t addr)
{
	int i;
	struct malo_cmdheader *hdr;

	hdr = sc->sc_cmd_mem;
	malo_ctl_write4(sc, 0x0c10, (uint32_t)addr);
	malo_ctl_barrier(sc, BUS_SPACE_BARRIER_WRITE);
	malo_ctl_write4(sc, 0x0c18, 2); /* CPU_TRANSFER_CMD */
	malo_ctl_barrier(sc, BUS_SPACE_BARRIER_WRITE);

	for (i = 0; i < 10; i++) {
		DELAY(100);
		bus_dmamap_sync(sc->sc_cmd_tag, sc->sc_cmd_map,
		    BUS_DMASYNC_POSTWRITE | BUS_DMASYNC_POSTREAD);
		if (hdr->cmd & htole16(0x8000))
			break;
	}

	if (i == 10)
		return (ETIMEDOUT);

	return (0);
}

static int
malo_init_rx_ring(struct malo_softc *sc)
{
	struct malo_rx_desc *desc;
	struct malo_rx_data *data;
	bus_addr_t addr;
	int i;

	MALO_LOCK_ASSERT(sc);

	sc->sc_rxring.cur = 0;
	sc->sc_rxring.next = 0;

	bzero(sc->sc_rxring.desc,
	    sizeof(struct malo_rx_desc) * MALO_RX_RING_COUNT);
	for (i = 0; i < MALO_RX_RING_COUNT; i++) {
		desc = &sc->sc_rxring.desc[i];
		data = &sc->sc_rxring.data[i];
		data->m = NULL;
		if (malo_newbuf(sc, i) != 0)
			return (ENOBUFS);
		if (i == (MALO_RX_RING_COUNT - 1))
			addr = MALO_RX_RING_ADDR(sc, 0);
		else
			addr = MALO_RX_RING_ADDR(sc, i + 1);
		desc->status = htole16(1);
		desc->physnext = htole32(addr);
	}

	bus_dmamap_sync(sc->sc_rxring.ring_tag, sc->sc_rxring.ring_map,
	    BUS_DMASYNC_PREREAD | BUS_DMASYNC_PREWRITE);

	return (0);
}

static int
malo_init_tx_ring(struct malo_softc *sc)
{
	struct malo_tx_desc *desc;
	struct malo_tx_data *data;
	bus_addr_t addr;
	int i;

	MALO_LOCK_ASSERT(sc);

	sc->sc_txring.queued = 0;
	sc->sc_txring.cur = sc->sc_txring.next = sc->sc_txring.stat = 0;

	bzero(sc->sc_txring.desc,
	    sizeof(struct malo_tx_desc) * MALO_RX_RING_COUNT);
	for (i = 0; i < MALO_TX_RING_COUNT; i++) {
		desc = &sc->sc_txring.desc[i];
		data = &sc->sc_txring.data[i];
		data->m = NULL;
		data->softstat = 0;
		data->ni = NULL;
		if (i == (MALO_TX_RING_COUNT - 1))
			addr = MALO_TX_RING_ADDR(sc, 0);
		else
			addr = MALO_TX_RING_ADDR(sc, i + 1);
		desc->physnext = htole32(addr);
	}

	bus_dmamap_sync(sc->sc_txring.ring_tag, sc->sc_txring.ring_map,
	    BUS_DMASYNC_PREREAD | BUS_DMASYNC_PREWRITE);

	return (0);
}

static __inline void
malo_enable_intr(struct malo_softc *sc)
{

	malo_ctl_write4(sc, 0x0c34, 0x1f);
	malo_ctl_barrier(sc, BUS_SPACE_BARRIER_WRITE);
	malo_ctl_write4(sc, 0x0c3c, 0x1f);
	malo_ctl_barrier(sc, BUS_SPACE_BARRIER_WRITE);
}

static __inline void
malo_disable_intr(struct malo_softc *sc)
{

	malo_ctl_read4(sc, 0x0c30);
	malo_ctl_write4(sc, 0x0c30, 0);
	malo_ctl_write4(sc, 0x0c34, 0);
	malo_ctl_write4(sc, 0x0c3c, 0);
}

static void
malo_init(void *xsc)
{
	struct malo_softc *sc;

	sc = xsc;
	MALO_LOCK(sc);
	malo_init_locked(sc);
	MALO_UNLOCK(sc);
}

static void
malo_init_locked(struct malo_softc *sc)
{
	struct ieee80211com *ic;
	struct ifnet *ifp;
	uint8_t chan;
	int error;

	MALO_LOCK_ASSERT(sc);

	DPRINTF(sc, "%s\n", __func__);

	ic = &sc->sc_ic;
	ifp = sc->sc_ifp;
	/* if interface already runs stop it first */
	if ((ifp->if_drv_flags & IFF_DRV_RUNNING) != 0)
		malo_stop(sc);

	/* disable interrupts */
	malo_disable_intr(sc);

	/* load firmware */
	/* XXX why firmware loading comes here... should be in device attach */
	if ((error = malo_load_bootimg(sc)))
		goto fail;
	if ((error = malo_load_firmware(sc)))
		goto fail;

	/* enable interrupts */
	malo_enable_intr(sc);

	/* XXX ??? */
	IEEE80211_ADDR_COPY(ic->ic_myaddr, IF_LLADDR(ifp));

	malo_init_tx_ring(sc);
	if ((error = malo_init_rx_ring(sc)) != 0) {
		device_printf(sc->sc_dev, "initializing Rx ring failed!\n");
		goto fail;
	}

	if ((error = malo_cmd_get_spec(sc)))
		goto fail;

	/* select default channel */
	ic->ic_bss->ni_chan = ic->ic_bsschan;
	chan = ieee80211_chan2ieee(ic, ic->ic_bss->ni_chan);

	/* initialize hardware */
	if ((error = malo_cmd_set_channel(sc, chan))) {
		device_printf(sc->sc_dev, "setting channel failed!\n");
		goto fail;
	}
	if ((error = malo_cmd_set_antenna(sc, 1))) {
		device_printf(sc->sc_dev, "setting RX antenna failed!\n");
		goto fail;
	}
	if ((error = malo_cmd_set_antenna(sc, 2))) {
		device_printf(sc->sc_dev, "setting TX antenna failed!\n");
		goto fail;
	}
	if ((error = malo_cmd_set_radio(sc, 1, 5))) {
		device_printf(sc->sc_dev, "turn radio on failed!\n");
		goto fail;
	}
	if ((error = malo_cmd_set_txpower(sc, 100))) {
		device_printf(sc->sc_dev, "setting TX power failed!\n");
		goto fail;
	}
	if ((error = malo_cmd_set_rts(sc, IEEE80211_RTS_MAX))) {
		device_printf(sc->sc_dev, "setting RTS failed!\n");
		goto fail;
	}

	/* WEP */
	if (sc->sc_ic.ic_flags & IEEE80211_F_PRIVACY) {
		/* set key */
		if ((error = malo_set_wepkey(sc))) {
			device_printf(sc->sc_dev, "setting WEP key failed!\n");
			goto fail;
		}
	}

	ifp->if_drv_flags |= IFF_DRV_RUNNING;
	ifp->if_drv_flags &= ~IFF_DRV_OACTIVE;
	ic->ic_state = IEEE80211_S_INIT;

	MALO_UNLOCK(sc);
	if (ic->ic_opmode != IEEE80211_M_MONITOR)
		/* start background scanning */
		ieee80211_new_state(ic, IEEE80211_S_SCAN, -1);
	else
		/* in monitor mode change directly into run state */
		ieee80211_new_state(ic, IEEE80211_S_RUN, -1);

	MALO_LOCK(sc);
	/* XXX kicking bg scanning here? */

	return;

fail:
	/* reset adapter */
	DPRINTF(sc, "%s failed, resetting card\n", __func__);
	malo_ctl_write4(sc, 0x0c18, (1 << 15));
}

static int
malo_ioctl(struct ifnet *ifp, u_long cmd, caddr_t data)
{
	struct malo_softc *sc;
	struct ieee80211com *ic;
	int error;
#if 0
	uint8_t chan;
#endif

	sc = ifp->if_softc;
	ic = &sc->sc_ic;
	error = 0;
	switch (cmd) {
	case SIOCSIFFLAGS:
		MALO_LOCK(sc);
		if (ifp->if_flags & IFF_UP) {
			if ((ifp->if_drv_flags & IFF_DRV_RUNNING) == 0)
				malo_init_locked(sc);
		} else {
			if ((ifp->if_drv_flags & IFF_DRV_RUNNING) != 0)
				malo_stop(sc);
		}
		MALO_UNLOCK(sc);
		break;
#if 0
	/* XXX NOT SURE */
	case SIOCS80211CHANNEL:
		/* allow fast channel switching in monitor mode */
		error = ieee80211_ioctl(ifp, cmd, data);
		if (error == ENETRESET &&
		    ic->ic_opmode == IEEE80211_M_MONITOR) {
			if ((ifp->if_flags & IFF_UP) != 0 &&
			    (ifp->if_drv_flags & IFF_DRV_RUNNING) != 0) {
				ic->ic_bss->ni_chan = ic->ic_ibss_chan;
				chan = ieee80211_chan2ieee(ic,
				    ic->ic_bss->ni_chan);
				malo_cmd_set_channel(sc, chan);
			}
			error = 0;
		}
		break;
#endif
	default:
		error = ieee80211_ioctl(ic, cmd, data);
		break;
	}

	if (error == ENETRESET) {
		if ((ifp->if_flags & IFF_UP) != 0 &&
		    (ifp->if_drv_flags & IFF_DRV_RUNNING) != 0)
			malo_init(ifp);
		error = 0;
	}

	return (error);
}

#if 0
static struct mbuf *
malo_fragment(struct mbuf *m0, int how, int length)
{
	struct mbuf *m_new = NULL, *m_final = NULL;
	int progress = 0;

	if (!(m0->m_flags & M_PKTHDR))
		return (m0);
	
	if ((length == 0) || (length < -2))
		return (m0);

	m_fixhdr(m0); /* Needed sanity check */

	m_final = m_getcl(how, MT_DATA, M_PKTHDR);

	if (m_final == NULL)
		goto nospace;

	if (m_dup_pkthdr(m_final, m0, how) == 0)
		goto nospace;

	m_new = m_final;

	if (length == -1)
		length = 1 + (arc4random() & 255);

	while (progress < m0->m_pkthdr.len) {
		int fraglen;

		if (length > 0)
			fraglen = length;
		else
			fraglen = 1 + (arc4random() & 255);
		if (fraglen > m0->m_pkthdr.len - progress)
			fraglen = m0->m_pkthdr.len - progress;

		if (fraglen > MCLBYTES)
			fraglen = MCLBYTES;

		if (m_new == NULL) {
			m_new = m_getcl(how, MT_DATA, 0);
			if (m_new == NULL)
				goto nospace;
		}

		m_copydata(m0, progress, fraglen, mtod(m_new, caddr_t));
		progress += fraglen;
		m_new->m_len = fraglen;
		if (m_new != m_final)
			m_cat(m_final, m_new);
		m_new = NULL;
	}
	m_freem(m0);
	m0 = m_final;
	return (m0);
nospace:
	if (m_final)
		m_freem(m_final);
	/* Return the original chain on failure */
	return (m0);
}
#endif

static void
malo_start(struct ifnet *ifp)
{
	struct malo_softc *sc;

	sc = ifp->if_softc;
	MALO_LOCK(sc);
	malo_start_locked(ifp);
	MALO_UNLOCK(sc);
}

static void
malo_start_locked(struct ifnet *ifp)
{
	struct malo_softc *sc;
	struct ieee80211com *ic;
	struct mbuf *m;
	struct ieee80211_node *ni;
	struct ether_header *eh;

	sc = ifp->if_softc;

	MALO_LOCK_ASSERT(sc);

	DPRINTFN(sc, 2, "%s\n", __func__);

	if ((ifp->if_drv_flags & (IFF_DRV_RUNNING | IFF_DRV_OACTIVE)) !=
	    IFF_DRV_RUNNING)
		return;

	ic = &sc->sc_ic;
	for (;;) {
		IF_DEQUEUE(&ic->ic_mgtq, m);
		if (m != NULL) {
			if (sc->sc_txring.queued >= MALO_TX_RING_COUNT - 1) {
				/* out of desc. */
				IF_PREPEND(&ic->ic_mgtq, m);
				ifp->if_drv_flags |= IFF_DRV_OACTIVE;
				break;
			}

			ni = (struct ieee80211_node *)m->m_pkthdr.rcvif;
			m->m_pkthdr.rcvif = NULL;
			if (bpf_peers_present(ic->ic_rawbpf))
				bpf_mtap(ic->ic_rawbpf, m);
			if (malo_tx_mgt(sc, &m, ni) != 0) {
				if (m != NULL) {
					/* out of desc. */
					IF_PREPEND(&ic->ic_mgtq, m);
					break;
				}
				ifp->if_oerrors++; /* XXX ??? */
				break;
			}
		} else {
			if (ic->ic_state != IEEE80211_S_RUN)
				break;
			IFQ_DRV_DEQUEUE(&ifp->if_snd, m);
			if (m == NULL)
				break;
			if (sc->sc_txring.queued >= MALO_TX_RING_COUNT - 1) {
				IFQ_DRV_PREPEND(&ifp->if_snd, m);
				ifp->if_drv_flags |= IFF_DRV_OACTIVE;
				break;
			}

			if (m->m_len < sizeof(struct ether_header) &&
			    (m = m_pullup(m,
			    sizeof(struct ether_header))) == NULL)
				continue;
			eh = mtod(m, struct ether_header *);
			ni = ieee80211_find_txnode(ic, eh->ether_dhost);
			if (ni == NULL) {
				m_freem(m);
				continue;
			}
			BPF_MTAP(ifp, m);
			m = ieee80211_encap(ic, m, ni);
			if (m == NULL) {
				ieee80211_free_node(ni);
				continue;
			}
			if (bpf_peers_present(ic->ic_rawbpf))
				bpf_mtap(ic->ic_rawbpf, m);
			if (malo_tx_data(sc, &m, ni) != 0) {
				ieee80211_free_node(ni);
				if (m != NULL) {
					/* out of desc. */
					IFQ_DRV_PREPEND(&ifp->if_snd, m);
					ifp->if_drv_flags |= IFF_DRV_OACTIVE;
					break;
				}
				ifp->if_oerrors++;
				break;
			}
		}
		/* XXX should have better way */
		ifp->if_timer = 5;
	}
}

static void
malo_stop(struct malo_softc *sc)
{
	struct ieee80211com *ic;
	struct ifnet *ifp;
	struct malo_tx_data *txd;
	struct malo_rx_data *rxd;
	struct malo_tx_desc *tx_desc;
	struct malo_rx_desc *rx_desc;
	int i;

	MALO_LOCK_ASSERT(sc);

	DPRINTF(sc, "%s\n", __func__);

	/*
	 * XXX should have better way to stop active Tx/Rx DMA
	 */

	ifp = sc->sc_ifp;
	ic = &sc->sc_ic;

	/* reset adapter */
	if ((ifp->if_drv_flags & IFF_DRV_RUNNING) != 0)
		malo_ctl_write4(sc, 0x0c18, (1 << 15));
	/* XXX it seems that it needs additional delay to stop F/W. */
	DELAY(1000);

	/* device is not running anymore */
	ifp->if_drv_flags &= ~(IFF_DRV_RUNNING | IFF_DRV_OACTIVE);
	ifp->if_timer = 0;

	/* change back to initial state */
	MALO_UNLOCK(sc);
	ieee80211_new_state(ic, IEEE80211_S_INIT, -1);
	MALO_LOCK(sc);

	/* reset RX / TX rings */
	/* XXX disable intr, disable active DMA inprogress, Drain Tx/Rx queue etc */
	/* Free Rx/Tx mbufs still in the queues */
	for (i = 0; i < MALO_RX_RING_COUNT; i++) {
		rx_desc = &sc->sc_rxring.desc[i];
		rxd = &sc->sc_rxring.data[i];
		if (rxd->m != NULL) {
			bus_dmamap_sync(sc->sc_rxring.rx_tag, rxd->map,
			    BUS_DMASYNC_POSTREAD);
			bus_dmamap_unload(sc->sc_rxring.rx_tag, rxd->map);
			m_freem(rxd->m);
			rxd->m = NULL;
		}
		rx_desc->status = 0;
	}
	bus_dmamap_sync(sc->sc_rxring.rx_tag, sc->sc_rxring.rx_map,
	    BUS_DMASYNC_PREREAD | BUS_DMASYNC_PREWRITE);
	sc->sc_rxring.cur = sc->sc_rxring.next = 0;

	for (i = 0; i < MALO_TX_RING_COUNT; i++) {
		tx_desc = &sc->sc_txring.desc[i];
		txd = &sc->sc_txring.data[i];
		if (txd->m != NULL) {
			bus_dmamap_sync(sc->sc_txring.tx_tag, txd->map,
			    BUS_DMASYNC_POSTWRITE);
			bus_dmamap_unload(sc->sc_txring.tx_tag, txd->map);
			m_freem(txd->m);
			txd->m = NULL;
		}
		/*
		 * The node has already been freed at that point so don't call
		 * ieee80211_release_node() here.
		 */
		txd->ni = NULL;

		tx_desc->status = 0;
	}
	bus_dmamap_sync(sc->sc_txring.tx_tag, sc->sc_txring.tx_map,
	    BUS_DMASYNC_PREREAD | BUS_DMASYNC_PREWRITE);
	sc->sc_txring.queued = 0;
	sc->sc_txring.cur = sc->sc_txring.next = sc->sc_txring.stat = 0;

	/* set initial rate */
	sc->sc_last_txrate = -1;
}

/* XXX should be rewritten */
static void
malo_watchdog(struct ifnet *ifp)
{
	struct malo_softc *sc;

	sc = ifp->if_softc;
	device_printf(sc->sc_dev, "XXX watchdog timeout!!!\n");
}

static void
malo_scan_start(struct ieee80211com *ic)
{
	struct malo_softc *sc;
	struct ifnet *ifp;

	ifp = ic->ic_ifp;
	sc = ifp->if_softc;
	MALO_LOCK(sc);

	if (malo_cmd_set_prescan(sc) != 0)
		DPRINTF(sc, "can't set prescan\n");
	MALO_UNLOCK(sc);
}

static void
malo_scan_end(struct ieee80211com *ic)
{
	struct malo_softc *sc;
	struct ifnet *ifp;

	ifp = ic->ic_ifp;
	sc = ifp->if_softc;
	MALO_LOCK(sc);

	if (malo_cmd_set_postscan(sc, ic->ic_myaddr, 1) != 0)
		DPRINTF(sc, "can't set postscan\n");
	if (malo_cmd_set_aid(sc, ic->ic_bss->ni_bssid,
	    ic->ic_bss->ni_associd) != 0)
		DPRINTF(sc, "can't set aid\n");
	MALO_UNLOCK(sc);
}

static void
malo_set_channel(struct ieee80211com *ic)
{
	struct malo_softc *sc;
	struct ifnet *ifp;
	uint8_t chan;

	ifp = ic->ic_ifp;
	sc = ifp->if_softc;

	MALO_LOCK(sc);
	chan = ieee80211_chan2ieee(ic, ic->ic_curchan);
	if (malo_cmd_set_channel(sc, chan) != 0)
		DPRINTF(sc, "can't set channel %u\n", chan);
	MALO_UNLOCK(sc);
}

static int
malo_newstate(struct ieee80211com *ic, enum ieee80211_state nstate, int arg)
{
	struct malo_softc *sc;
	struct ifnet *ifp;
	enum ieee80211_state ostate;
#if 0
	uint8_t chan;
#endif
	int error, rate;

	ifp = ic->ic_ifp;
	sc = ifp->if_softc;

	/* XXX */
	MALO_LOCK(sc);

	DPRINTF(sc, "%s: %s -> %s\n", __func__,
	    ieee80211_state_name[ic->ic_state],
	    ieee80211_state_name[nstate]);

	ostate = ic->ic_state;
	callout_stop(&sc->sc_scan_to);

	switch (nstate) {
	case IEEE80211_S_INIT:
		/* XXX Should disable Rx interrupt */
		break;
	case IEEE80211_S_SCAN:
		/* XXX NOT SURE */
#if 0
		if (ostate == IEEE80211_S_INIT) {
			if (malo_cmd_set_prescan(sc) != 0)
				DPRINTF(sc, "can't set prescan\n");
		} else {
			chan = ieee80211_chan2ieee(ic, ic->ic_curchan);
			malo_cmd_set_channel(sc, chan);
		}
		callout_reset(&sc->sc_scan_to, hz / 2, malo_next_scan, sc);
#endif
		break;
	case IEEE80211_S_AUTH:
		/* XXX NOT SURE */
#if 0
		malo_cmd_set_postscan(sc, ic->ic_myaddr, 1);
		chan = ieee80211_chan2ieee(ic, ic->ic_bss->ni_chan);
		malo_cmd_set_channel(sc, chan);
#endif
		break;
	case IEEE80211_S_ASSOC:
		if (ic->ic_flags & IEEE80211_F_SHPREAMBLE)
			malo_cmd_set_radio(sc, 1, 3); /* short preamble */
		else
			malo_cmd_set_radio(sc, 1, 1); /* long preamble */

		malo_cmd_set_aid(sc, ic->ic_bss->ni_bssid,
		    ic->ic_bss->ni_associd);

		if (ic->ic_fixed_rate == -1)
			/* automatic rate adaption */
			malo_cmd_set_rate(sc, 0);
		else {
			/* fixed rate */
			rate = malo_fix2rate(ic->ic_fixed_rate);
			malo_cmd_set_rate(sc, rate);
		}

		malo_set_slot(sc);
		break;
	default:
		break;
	}

	MALO_UNLOCK(sc);

	/* Invoke the parent method to complete the work */
	error = (sc->sc_newstate(ic, nstate, arg));
	/* Start any timers */

	return (error);
}

/* XXX */
static void
malo_newassoc(struct ieee80211_node *ni, int isnew)
{

}

static struct ieee80211_node *
malo_node_alloc(struct ieee80211_node_table *nt)
{
	struct ieee80211com *ic;
	struct malo_node *wn;

	ic = nt->nt_ic;
	wn = malloc(sizeof(struct malo_node), M_80211_NODE, M_NOWAIT | M_ZERO);
	if (wn == NULL)
		return (NULL);

	return ((struct ieee80211_node *)wn);
}

static int
malo_media_change(struct ifnet *ifp)
{
	int error;

	DPRINTF((struct malo_softc *)(ifp->if_softc), "%s\n", __func__);

	error = ieee80211_media_change(ifp);
	if (error != ENETRESET)
		return (error);

	if ((ifp->if_flags & IFF_UP) != 0 &&
	    (ifp->if_drv_flags & IFF_DRV_RUNNING) != 0)
		malo_init(ifp);

	return (0);
}

static void
malo_media_status(struct ifnet *ifp, struct ifmediareq *imr)
{
	struct malo_softc *sc;
	struct ieee80211com *ic;

	sc = ifp->if_softc;
	ic = &sc->sc_ic;
	imr->ifm_status = IFM_AVALID;
	imr->ifm_active = IFM_IEEE80211;
	if (ic->ic_state == IEEE80211_S_RUN)
		imr->ifm_status |= IFM_ACTIVE;

	/* XXX unlocked access */
	/* report last TX rate used by chip */
	imr->ifm_active |= ieee80211_rate2media(ic, sc->sc_last_txrate,
	    ic->ic_curmode);

	switch (ic->ic_opmode) {
	case IEEE80211_M_STA:
		break;
	case IEEE80211_M_IBSS:
		imr->ifm_active |= IFM_IEEE80211_ADHOC;
		break;
	case IEEE80211_M_MONITOR:
		imr->ifm_active |= IFM_IEEE80211_MONITOR;
		break;
	case IEEE80211_M_AHDEMO:
		break;
	case IEEE80211_M_HOSTAP:
		break;
	default:
		break;
	}

	switch (ic->ic_curmode) {
		case IEEE80211_MODE_11B:
			imr->ifm_active |= IFM_IEEE80211_11B;
			break;
		case IEEE80211_MODE_11G:
			imr->ifm_active |= IFM_IEEE80211_11G;
			break;
	}
}

static int
malo_chip2rate(int chip_rate)
{

	switch (chip_rate) {
	/* CCK rates */
	case  0:	return (2);
	case  1:	return (4);
	case  2:	return (11);
	case  3:	return (22);

	/* OFDM rates */
	case  4:	return (0); /* reserved */
	case  5:	return (12);
	case  6:	return (18);
	case  7:	return (24);
	case  8:	return (36);
	case  9:	return (48);
	case 10:	return (72);
	case 11:	return (96);
	case 12:	return (108);

	/* no rate select yet or unknown rate */
	default:	return (-1);
	}
}

static int
malo_fix2rate(int fix_rate)
{

	switch (fix_rate) {
	/* CCK rates */
	case  0:	return (2);
	case  1:	return (4);
	case  2:	return (11);
	case  3:	return (22);

	/* OFDM rates */
	case  4:	return (12);
	case  5:	return (18);
	case  6:	return (24);
	case  7:	return (36);
	case  8:	return (48);
	case  9:	return (72);
	case 10:	return (96);
	case 11:	return (108);

	/* unknown rate: should not happen */
	default:	return (0);
	}
}

/* XXX SCAN related one should be rewritten */
#if 0
static void
malo_next_scan(void *arg)
{
	struct malo_softc *sc;
	struct ieee80211com *ic;
	struct ifnet *ifp;

	sc = arg;
	MALO_LOCK_ASSERT(sc);

	ic = &sc->sc_ic;
	ifp = sc->sc_ifp;
	DPRINTF(sc, "%s\n", __func__);

	if (ic->ic_state == IEEE80211_S_SCAN)
		ieee80211_next_scan(ifp);
}
#endif

static void
malo_tx_intr(struct malo_softc *sc)
{
	struct ieee80211com *ic;
	struct ifnet *ifp;
	struct malo_tx_desc *desc;
	struct malo_tx_data *data;
	struct malo_node *rn;
	int stat, prog;

	MALO_LOCK_ASSERT(sc);

	DPRINTFN(sc, 2, "%s\n", __func__);

	if (sc->sc_txring.queued == 0)
		return;

	bus_dmamap_sync(sc->sc_txring.ring_tag, sc->sc_txring.ring_map,
	    BUS_DMASYNC_POSTREAD | BUS_DMASYNC_POSTWRITE);

	ifp = sc->sc_ifp;
	ic = &sc->sc_ic;
	prog = 0;
	for (stat = sc->sc_txring.stat; stat != sc->sc_txring.cur;
	    MALO_TX_RING_INC(stat)) {
		desc = &sc->sc_txring.desc[stat];
		data = &sc->sc_txring.data[stat];
		rn = (struct malo_node *)data->ni;

		/* check if TX descriptor is not owned by FW anymore */
		if ((le32toh(desc->status) & 0x80000000) ||
		    !(le32toh(data->softstat) & 0x80))
			break;

		/* if no frame has been sent, ignore */
		if (rn == NULL)
			continue;

		prog++;
		/* check TX state */
		switch (le32toh(desc->status) & 0x1) {
		case 0x1:
			DPRINTFN(sc, 2, "data frame was sent successfully\n");
			ifp->if_opackets++;
			break;
		default:
			DPRINTF(sc, "data frame sending error\n");
			ifp->if_oerrors++;
			break;
		}

		/* save last used TX rate */
		sc->sc_last_txrate = malo_chip2rate(desc->datarate);

		/* XXX Should check the last mbuf chain!!!! */
		/* printf("TX INTR: %08x \n", le32toh(desc->status));*/

		/* cleanup TX data and TX descritpor */
		bus_dmamap_sync(sc->sc_txring.tx_tag, data->map,
		    BUS_DMASYNC_POSTWRITE);
		bus_dmamap_unload(sc->sc_txring.tx_tag, data->map);

		KASSERT(data->m != NULL, ("%s: freeing NULL mbuf!", __func__));

		m_freem(data->m);
		ieee80211_free_node(data->ni);
		data->m = NULL;
		data->ni = NULL;
		data->softstat &= htole32(~0x80);
		desc->status = 0;
		desc->len = 0;

		DPRINTFN(sc, 2, "tx done idx=%u\n", prog);

		sc->sc_txring.queued--;
	}

	if (prog > 0) {
		sc->sc_txring.stat = stat;
		if (sc->sc_txring.queued == 0) {
			ifp->if_timer = 0;
			sc->sc_tx_timer = 0;
		}
		ifp->if_drv_flags &= ~IFF_DRV_OACTIVE;

		bus_dmamap_sync(sc->sc_txring.ring_tag, sc->sc_txring.ring_map,
		    BUS_DMASYNC_PREREAD | BUS_DMASYNC_PREWRITE);
	}
	malo_start_locked(ifp);
}

static int
malo_tx_mgt(struct malo_softc *sc, struct mbuf **m_head, struct ieee80211_node *ni)
{
	struct ieee80211com *ic;
	struct ifnet *ifp;
	struct malo_tx_desc *desc;
	struct malo_tx_data *data, *data_last;
	struct ieee80211_frame *wh;
	struct ieee80211_key *k;
	bus_dma_segment_t txsegs[MALO_MAX_SCATTER];
	bus_dmamap_t map;
	struct mbuf *m;
	int error, i, nsegs, prod, si;

	MALO_LOCK_ASSERT(sc);

	DPRINTFN(sc, 2, "%s\n", __func__);

	ic = &sc->sc_ic;
	ifp = sc->sc_ifp;

	if ((*m_head)->m_len < sizeof(struct ieee80211_frame)) {
		*m_head = m_pullup(*m_head, sizeof(struct ieee80211_frame));
		if (*m_head == NULL) {
			/* XXX I don't think the following is needed */
			ifp->if_ierrors++;
			return (ENOBUFS);
		}
	}
	wh = mtod(*m_head, struct ieee80211_frame *);

	if (wh->i_fc[1] & IEEE80211_FC1_WEP) {
		k = ieee80211_crypto_encap(ic, ni, *m_head);
		if (k == NULL) {
			m_freem(*m_head);
			*m_head = NULL;
			/* XXX better error code? */
			return (EIO);
		}

		/* packet header may have moved, reset our local pointer */
		wh = mtod(*m_head, struct ieee80211_frame *);
	}

	if (bpf_peers_present(sc->sc_drvbpf)) {
		struct malo_tx_radiotap_hdr *tap;

		tap = &sc->sc_txtap;
		tap->wt_flags = 0;
		tap->wt_rate = sc->sc_last_txrate;
		tap->wt_chan_freq = htole16(ic->ic_bss->ni_chan->ic_freq);
		tap->wt_chan_flags = htole16(ic->ic_bss->ni_chan->ic_flags);

		bpf_mtap2(sc->sc_drvbpf, tap, sc->sc_txtap_len, *m_head);
	}

	/*
	 * inject FW specific fields into the 802.11 frame
	 *
	 *  2 bytes FW len (inject)
	 * 24 bytes 802.11 frame header
	 *  6 bytes addr4 (inject)
	 *  n bytes 802.11 frame body
	 */
	/* XXX check for writability of the mbuf? */
	if (M_LEADINGSPACE(*m_head) < 8) {
		if (M_TRAILINGSPACE(*m_head) < 8)
			panic("%s: not enough space for mbuf dance",
			    ifp->if_xname);
		bcopy((*m_head)->m_data, (*m_head)->m_data + 8,
		    (*m_head)->m_len);
		(*m_head)->m_data += 8;
	}

	/* XXX should eliminate this copy, assume single mbuf chain */
	/* move frame header */
	bcopy((*m_head)->m_data, (*m_head)->m_data - 6, sizeof(*wh));
	(*m_head)->m_data -= 8;
	(*m_head)->m_len += 8;
	(*m_head)->m_pkthdr.len += 8;
	/* FW len */
	*mtod(*m_head, uint16_t *) = htole16((*m_head)->m_len - 32);

	data_last = &sc->sc_txring.data[sc->sc_txring.cur];
	data = data_last;
	map = data_last->map;

	error = bus_dmamap_load_mbuf_sg(sc->sc_txring.tx_tag, map,
	    *m_head, txsegs, &nsegs, 0);
	if (error == EFBIG) {
		m = m_defrag(*m_head, M_DONTWAIT);
		if (m == NULL) {
			m_freem(*m_head);
			*m_head = NULL;
			return (ENOBUFS);
		}
		*m_head = m;
		error = bus_dmamap_load_mbuf_sg(sc->sc_txring.tx_tag, map,
		    *m_head, txsegs, &nsegs, 0);
		if (error != 0) {
			m_freem(*m_head);
			*m_head = NULL;
			return (ENOBUFS);
		}
	} else if (error != 0) {
		/* XXX Since we've modified origial mbuf, drop it */
		m_freem(*m_head);
		*m_head = NULL;
		return (error);
	}
	if (nsegs == 0) {
		m_freem(*m_head);
		*m_head = NULL;
		return (EIO);
	}
	if (sc->sc_txring.queued + nsegs >= MALO_TX_RING_COUNT - 1) {
		bus_dmamap_unload(sc->sc_txring.tx_tag, map);
		/* XXX Since we've modified origial mbuf, drop it */
		m_freem(*m_head);
		*m_head = NULL;
		return (ENOBUFS);
	}

	bus_dmamap_sync(sc->sc_txring.tx_tag, map, BUS_DMASYNC_PREWRITE);

	si = prod = sc->sc_txring.cur;
	for (i = 0; i < nsegs; i++) {
		desc = &sc->sc_txring.desc[prod];
		desc->len = htole16(txsegs[i].ds_len);
		desc->datarate = 0; /* 0 = mgmt frame, 1 = data frame */
		desc->physdata = htole32(txsegs[i].ds_addr);
		if (i == 0)
			desc->status = htole32(0/*0x00000001*/);
		else
			desc->status = htole32(0/*0x00000001*/ | 0x80000000);
		sc->sc_txring.queued++;
		MALO_TX_RING_INC(prod);
	}
	sc->sc_txring.cur = prod;

	/* XXX set EOP */
	prod = (prod + MALO_TX_RING_COUNT - 1) % MALO_TX_RING_COUNT;
	desc = &sc->sc_txring.desc[prod];
	desc->status |= htole32(1/*0*/);	/* XXX FIND EOP mark!!! */

	/* turn the first descriptor ownership to hardware */
	desc = &sc->sc_txring.desc[si];
	desc->status |= htole32(0x80000000);
	
	data = &sc->sc_txring.data[prod];
	map = data_last->map;
	data_last->map = data->map;
	data->map = map;
	data->m = *m_head;
	data->ni = ni;
	data->softstat |= htole32(0x80);

	/* Sync descriptors. */
	bus_dmamap_sync(sc->sc_txring.ring_tag, sc->sc_txring.ring_map,
	    BUS_DMASYNC_PREREAD | BUS_DMASYNC_PREWRITE);

	DPRINTFN(sc, 2, "sending mgmt frame, pktlen=%u, idx=%u\n",
	    (*m_head)->m_pkthdr.len, sc->sc_txring.cur);

	/* kick mgmt TX */
	malo_ctl_write4(sc, 0x0c18, 1);
	malo_ctl_barrier(sc, BUS_SPACE_BARRIER_WRITE);

	return (0);
}

static int
malo_tx_data(struct malo_softc *sc, struct mbuf **m_head,
    struct ieee80211_node *ni)
{
	struct ieee80211com *ic;
	struct ifnet *ifp;
	struct malo_tx_desc *desc;
	struct malo_tx_data *data, *data_last;
	struct ieee80211_frame *wh;
	struct ieee80211_key *k;
	bus_dma_segment_t txsegs[MALO_MAX_SCATTER];
	bus_dmamap_t map;
	struct mbuf *m, *mnew;
	int error, i, nsegs, prod, si;

	MALO_LOCK_ASSERT(sc);

	DPRINTFN(sc, 2, "%s\n", __func__);

	ic = &sc->sc_ic;
	ifp = sc->sc_ifp;

	if ((*m_head)->m_len < sizeof(struct ieee80211_frame)) {
		*m_head = m_pullup(*m_head, sizeof(struct ieee80211_frame));
		if (*m_head == NULL) {
			/* XXX I don't think the following is needed */
			ifp->if_ierrors++;
			return (ENOBUFS);
		}
	}
	wh = mtod(*m_head, struct ieee80211_frame *);

	if (wh->i_fc[1] & IEEE80211_FC1_WEP) {
		k = ieee80211_crypto_encap(ic, ni, *m_head);
		if (k == NULL) {
			m_freem(*m_head);
			*m_head = NULL;
			return (ENOBUFS);
		}

		/* packet header may have moved, reset our local pointer */
		wh = mtod(*m_head, struct ieee80211_frame *);
	}

	if (bpf_peers_present(sc->sc_drvbpf)) {
		struct malo_tx_radiotap_hdr *tap;

		tap = &sc->sc_txtap;
		tap->wt_flags = 0;
		tap->wt_rate = sc->sc_last_txrate;
		tap->wt_chan_freq = htole16(ic->ic_bss->ni_chan->ic_freq);
		tap->wt_chan_flags = htole16(ic->ic_bss->ni_chan->ic_flags);

		bpf_mtap2(sc->sc_drvbpf, tap, sc->sc_txtap_len, *m_head);
	}

	/*
	 * inject FW specific fields into the 802.11 frame
	 *
	 *  2 bytes FW len (inject)
	 * 24 bytes 802.11 frame header
	 *  6 bytes addr4 (inject)
	 *  n bytes 802.11 frame body
	 *
	 * For now copy all into a new mcluster.
	 */
	/* XXX This should be rewritten! No need to copy/alloc cluster! */
	mnew = m_getcl(M_DONTWAIT, MT_DATA, M_PKTHDR);
	if (mnew == NULL)
		return (ENOBUFS);

	/* XXX This nullify DMA op!!! */
	*mtod(mnew, uint16_t *) = htole16((*m_head)->m_pkthdr.len - 24); /* FW len */
	bcopy(wh, mtod(mnew, caddr_t) + 2, sizeof(*wh));
	bzero(mtod(mnew, caddr_t) + 26, 6);
	m_copydata(*m_head, sizeof(*wh), (*m_head)->m_pkthdr.len - sizeof(*wh),
	    mtod(mnew, caddr_t) + 32);
	mnew->m_pkthdr.len = mnew->m_len = (*m_head)->m_pkthdr.len + 8;
	m_freem(*m_head);
	*m_head = mnew;

#if 0
	*m_head = malo_fragment(mnew, M_DONTWAIT, 60);
	if (*m_head == mnew) {
		printf("m_fragment fail\n");
	} else {
		struct mbuf *n;
		int i;
		printf("m_fragment : ");
		for (n = *m_head, i = 0; n; n = n->m_next, i++)
			printf("%d ", n->m_len);
		printf(": %d fragmented\n", i);	
	}
#endif
	data_last = &sc->sc_txring.data[sc->sc_txring.cur];
	data = data_last;
	map = data_last->map;

	error = bus_dmamap_load_mbuf_sg(sc->sc_txring.tx_tag, map,
	    *m_head, txsegs, &nsegs, 0);
	if (error == EFBIG) {
		m = m_defrag(*m_head, M_DONTWAIT);
		if (m == NULL) {
			m_freem(*m_head);
			*m_head = NULL;
			return (ENOBUFS);
		}
		*m_head = m;
		error = bus_dmamap_load_mbuf_sg(sc->sc_txring.tx_tag, map,
		    *m_head, txsegs, &nsegs, 0);
		if (error != 0) {
			m_freem(*m_head);
			*m_head = NULL;
			return (ENOBUFS);
		}
	} else if (error != 0) {
		/* XXX Since we've modified origial mbuf, drop it */
		m_freem(*m_head);
		*m_head = NULL;
		return (error);
	}
	if (nsegs == 0) {
		m_freem(*m_head);
		*m_head = NULL;
		return (EIO);
	}
	if (sc->sc_txring.queued + nsegs >= MALO_TX_RING_COUNT - 1) {
		bus_dmamap_unload(sc->sc_txring.tx_tag, map);
		/* XXX Since we've modified origial mbuf, drop it */
		m_freem(*m_head);
		*m_head = NULL;
		return (ENOBUFS);
	}

	bus_dmamap_sync(sc->sc_txring.tx_tag, map, BUS_DMASYNC_PREWRITE);

	si = prod = sc->sc_txring.cur;
	for (i = 0; i < nsegs; i++) {
		desc = &sc->sc_txring.desc[prod];
		desc->len = htole16(txsegs[i].ds_len);
		desc->datarate = 1; /* 0 = mgmt frame, 1 = data frame */
		desc->physdata = htole32(txsegs[i].ds_addr);
		if (i == 0)
			desc->status = htole32(0x00000001);
		else
			desc->status = htole32(0x00000001 | 0x80000000);
		sc->sc_txring.queued++;
		MALO_TX_RING_INC(prod);
	}
	sc->sc_txring.cur = prod;

	/* XXX set EOP */
	prod = (prod + MALO_TX_RING_COUNT - 1) % MALO_TX_RING_COUNT;
	desc = &sc->sc_txring.desc[prod];
	desc->status |= htole32(0);	/* XXX FIND EOP mark!!! */

	/* turn the first descriptor ownership to hardware */
	desc = &sc->sc_txring.desc[si];
	desc->status |= htole32(0x80000000);
	
	data = &sc->sc_txring.data[prod];
	map = data_last->map;
	data_last->map = data->map;
	data->map = map;
	data->m = *m_head;
	data->ni = ni;
	data->softstat |= htole32(0x80);

	/* Sync descriptors. */
	bus_dmamap_sync(sc->sc_txring.ring_tag, sc->sc_txring.ring_map,
	    BUS_DMASYNC_PREREAD | BUS_DMASYNC_PREWRITE);

	DPRINTFN(sc, 2, "sending data frame, pktlen=%u, idx=%u\n",
	    (*m_head)->m_pkthdr.len, sc->sc_txring.cur);

	/* kick data TX */
	malo_ctl_write4(sc, 0x0c18, 1);
	malo_ctl_barrier(sc, BUS_SPACE_BARRIER_WRITE);

	return (0);
}

static __inline void
malo_discard_rxbuf(struct malo_softc *sc, int idx)
{
	struct malo_rx_desc *desc;
	struct malo_rx_data *data;

	data = &sc->sc_rxring.data[idx];
	desc = &sc->sc_rxring.desc[idx];
	/* XXX desc->len = data->m->m_len; */
	desc->rxctrl = 0;
}

static int
malo_newbuf(struct malo_softc *sc, int idx)
{
	struct malo_rx_desc *desc;
	struct malo_rx_data *data;
	struct mbuf *m;
	bus_dma_segment_t segs[1];
	bus_dmamap_t map;
	int nsegs;

	m = m_getcl(M_DONTWAIT, MT_DATA, M_PKTHDR);
	if (m == NULL)
		return (ENOBUFS);
	m->m_len = m->m_pkthdr.len = MCLBYTES;
	m_adj(m, ETHER_ALIGN);

	if (bus_dmamap_load_mbuf_sg(sc->sc_rxring.rx_tag,
	    sc->sc_rxring.rx_sparemap, m, segs, &nsegs, 0) != 0) {
		m_freem(m);
		return (ENOBUFS);
	}
	KASSERT(nsegs == 1, ("%s: %d segments returned!", __func__, nsegs));

	data = &sc->sc_rxring.data[idx];
	if (data->m != NULL) {
		bus_dmamap_sync(sc->sc_rxring.rx_tag, data->map,
		    BUS_DMASYNC_POSTREAD);
		bus_dmamap_unload(sc->sc_rxring.rx_tag, data->map);
	}
	map = data->map;
	data->map = sc->sc_rxring.rx_sparemap;
	sc->sc_rxring.rx_sparemap = map;
	bus_dmamap_sync(sc->sc_rxring.rx_tag, data->map, BUS_DMASYNC_PREREAD);
	data->m = m;

	desc = &sc->sc_rxring.desc[idx];
	desc->physdata = htole32(segs[0].ds_addr);
	/* XXX desc->len = htole32(segs[0].ds_len); */
	desc->rxctrl = 0;

	return (0);
}

static void
malo_rx_intr(struct malo_softc *sc)
{
	struct ieee80211com *ic;
	struct ifnet *ifp;
	struct malo_rx_desc *desc;
	struct malo_rx_data *data;
	struct ieee80211_frame *wh;
	struct ieee80211_node *ni;
	struct mbuf *m;
	uint32_t rxRdPtr, rxWrPtr;
	uint16_t len;
	uint8_t rssi;
	int i, cons;

	MALO_LOCK_ASSERT(sc);

	rxRdPtr = malo_mem_read4(sc, sc->sc_RxPdRdPtr);
	rxWrPtr = malo_mem_read4(sc, sc->sc_RxPdWrPtr);
	if (rxRdPtr == rxWrPtr) 
		return;

	bus_dmamap_sync(sc->sc_rxring.ring_tag, sc->sc_rxring.ring_map,
	    BUS_DMASYNC_POSTREAD | BUS_DMASYNC_POSTWRITE);

	ic = &sc->sc_ic;
	ifp = sc->sc_ifp;
	cons = sc->sc_rxring.cur;

	for (i = 0; i < MALO_RX_RING_COUNT && rxRdPtr != rxWrPtr; i++,
	    MALO_RX_RING_INC(cons)) {
		desc = &sc->sc_rxring.desc[cons];
		data = &sc->sc_rxring.data[cons];

		DPRINTFN(sc, 3, "rx intr idx=%d, rxctrl=0x%02x, rssi=%d, "
		    "status=0x%02x, channel=%d, len=%d, res1=%02x, rate=%d, "
		    "physdata=0x%04x, physnext=0x%04x, qosctrl=%02x, res2=%d\n",
		    cons, desc->rxctrl, desc->rssi, desc->status,
		    desc->channel, le16toh(desc->len), desc->reserved1,
		    desc->datarate, le32toh(desc->physdata),
		    le32toh(desc->physnext), desc->qosctrl, desc->reserved2);

		if ((desc->rxctrl & 0x80) == 0)
			break;

		rxRdPtr = le32toh(desc->physnext);

		/* XXX lack of Recevied packet error checking!!! */
		len = le16toh(desc->len);
		rssi = desc->rssi;
		m = data->m;

		if (malo_newbuf(sc, cons) != 0) {
			ifp->if_iqdrops++;
			malo_discard_rxbuf(sc, cons);
			continue;
		}

		/* finalize mbuf */
		m->m_pkthdr.rcvif = ifp;
		m->m_pkthdr.len = m->m_len = len;

		/*
		 * cut out FW specific fields from the 802.11 frame
		 *
		 *  2 bytes FW len (cut out)
		 * 24 bytes 802.11 frame header
		 *  6 bytes addr4 (cut out)
		 *  n bytes 802.11 frame data
		 */
		bcopy(m->m_data, m->m_data + 6, 26);
		m_adj(m, 8);

		if (bpf_peers_present(sc->sc_drvbpf)) {
			struct malo_rx_radiotap_hdr *tap;

			tap = &sc->sc_rxtap;
			tap->wr_flags = 0;
			tap->wr_chan_freq =
			    htole16(ic->ic_bss->ni_chan->ic_freq);
			tap->wr_chan_flags =
			    htole16(ic->ic_bss->ni_chan->ic_flags);
			tap->wr_rssi = rssi;
#if 0
			tap->wr_max_rssi = ic->ic_max_rssi;
#endif

			bpf_mtap2(sc->sc_drvbpf, tap, sc->sc_rxtap_len, m);
		}

		wh = mtod(m, struct ieee80211_frame *);

		MALO_UNLOCK(sc);

		/* XXX casting needed?? */
		ni = ieee80211_find_rxnode(ic, (struct ieee80211_frame_min *)wh);
		/* send the frame to the 802.11 layer */
		ieee80211_input(ic, m, ni, rssi, -95 /* XXX */, 0);

		/* node is no longer needed */
		ieee80211_free_node(ni);

		MALO_LOCK(sc);
	}

	malo_mem_write4(sc, sc->sc_RxPdRdPtr, rxRdPtr);
	sc->sc_rxring.cur = cons;

	bus_dmamap_sync(sc->sc_rxring.ring_tag, sc->sc_rxring.ring_map,
	    BUS_DMASYNC_PREREAD | BUS_DMASYNC_PREWRITE);

	/*
	 * In HostAP mode, ieee80211_input() will enqueue packets in if_snd
	 * without calling if_start().
	 */
	if (!IFQ_DRV_IS_EMPTY(&ifp->if_snd) &&
	    (ifp->if_drv_flags & IFF_DRV_OACTIVE) == 0)
		malo_start_locked(ifp);
}

static int
malo_load_bootimg(struct malo_softc *sc)
{
	char *name = "malo8335-h";
	const struct firmware *fw;
	int i;

	MALO_LOCK_ASSERT(sc);

	MALO_UNLOCK(sc);
	/* load boot firmware */
	fw = firmware_get(name);
	MALO_LOCK(sc);
	if (fw == NULL) {
		device_printf(sc->sc_dev,
		    "could not read microcode %s!\n", name);
		return (EIO);
	}

	/*
	 * It seems we are putting this code directly onto the stack of
	 * the ARM cpu. I don't know why we need to instruct the DMA
	 * engine to move the code. This is a big riddle without docu.
	 */
	DPRINTF(sc, "loading boot firmware version %u\n", fw->version);
	malo_mem_write2(sc, 0xbef8, 0x001);
	malo_mem_write2(sc, 0xbefa, fw->datasize);
	malo_mem_write4(sc, 0xbefc, 0);

	bus_write_region_1(sc->sc_res_mem[0], 0xbf00, fw->data, fw->datasize);

	/*
	 * we loaded the firmware into card memory now tell the CPU
	 * to fetch the code and execute it. The memory mapped via the
	 * first bar is internaly mapped to 0xc0000000.
	 */
	malo_send_cmd(sc, 0xc000bef8);

	/* wait for the device to go into FW loading mode */
	for (i = 0; i < 10; i++) {
		DELAY(50);
		malo_ctl_barrier(sc, BUS_SPACE_BARRIER_READ);
		if (malo_ctl_read4(sc, 0x0c14) == 0x5)
			break;
	}

	firmware_put(fw, FIRMWARE_UNLOAD);

	if (i == 10) {
		device_printf(sc->sc_dev, "timeout at boot firmware load!\n");
		return (ETIMEDOUT);
	}

	/* tell the card we're done and... */
	malo_mem_write2(sc, 0xbef8, 0x001);
	malo_mem_write2(sc, 0xbefa, 0);
	malo_mem_write4(sc, 0xbefc, 0);
	malo_send_cmd(sc, 0xc000bef8);

	DPRINTF(sc, "boot firmware loaded\n");

	return (0);
}

static int
malo_load_firmware(struct malo_softc *sc)
{
	struct malo_cmdheader *hdr;
	char *name = "malo8335-m";
	const struct firmware *fw;
	void *data;
	size_t count, bsize;
	int i, sn;

	MALO_LOCK_ASSERT(sc);

	MALO_UNLOCK(sc);
	/* load real firmware now */
	fw = firmware_get(name);
	MALO_LOCK(sc);
	if (fw == NULL) {
		device_printf(sc->sc_dev,
		    "could not read microcode %s!\n", name);
		return (EIO);
	}

	DPRINTF(sc, "uploading firmware\n");

	hdr = sc->sc_cmd_mem;
	data = hdr + 1;
	sn = 1;
	for (count = 0; count < fw->datasize; count += bsize) {
		bsize = MIN(256, fw->datasize - count);

		hdr->cmd = htole16(0x0001);
		hdr->size = htole16(bsize);
		hdr->seqnum = htole16(sn++);
		hdr->result = 0;

		bcopy((const char *)fw->data + count, data, bsize);

		bus_dmamap_sync(sc->sc_cmd_tag, sc->sc_cmd_map,
		    BUS_DMASYNC_PREWRITE);
		malo_send_cmd(sc, sc->sc_cmd_dmaaddr);
		bus_dmamap_sync(sc->sc_cmd_tag, sc->sc_cmd_map,
		    BUS_DMASYNC_POSTWRITE);
		DELAY(500);
	}
	firmware_put(fw, FIRMWARE_UNLOAD);

	DPRINTF(sc, "firmware upload finished\n");

	/*
	 * send a command with size 0 to tell that the firmware has been
	 * uploaded
	 */
	hdr->cmd = htole16(0x0001);
	hdr->size = 0;
	hdr->seqnum = htole16(sn++);
	hdr->result = 0;

	bus_dmamap_sync(sc->sc_cmd_tag, sc->sc_cmd_map, BUS_DMASYNC_PREWRITE);
	malo_send_cmd(sc, sc->sc_cmd_dmaaddr);
	bus_dmamap_sync(sc->sc_cmd_tag, sc->sc_cmd_map, BUS_DMASYNC_POSTWRITE);
	DELAY(100);

	DPRINTF(sc, "loading firmware\n");

	/* wait until firmware has been loaded */
	for (i = 0; i < 200; i++) {
		malo_ctl_write4(sc, 0x0c10, 0x5a);
		DELAY(500);
		malo_ctl_barrier(sc, BUS_SPACE_BARRIER_WRITE |
		     BUS_SPACE_BARRIER_READ);
		if (malo_ctl_read4(sc, 0x0c14) == 0xf0f1f2f4)
			break;
	}
	if (i == 200) {
		device_printf(sc->sc_dev, "timeout at firmware load!\n");
		return (ETIMEDOUT);
	}

	DPRINTF(sc, "firmware loaded\n");

	return (0);
}

static int
malo_set_wepkey(struct malo_softc *sc)
{
	struct ieee80211com *ic;
	struct ieee80211_key *k;
	int i;

	ic = &sc->sc_ic;
	k = ic->ic_crypto.cs_nw_keys;
	for (i = 0; i < IEEE80211_WEP_NKID; i++, k++) {
		if (k->wk_keylen == 0)
			continue;
		if (malo_cmd_set_wepkey(sc, k, i))
			return (ENXIO);
	}

	return (0);
}

static int
malo_set_slot(struct malo_softc *sc)
{
	struct ieee80211com *ic;

	MALO_LOCK_ASSERT(sc);

	ic = &sc->sc_ic;
	if (ic->ic_flags & IEEE80211_F_SHSLOT) {
		/* set short slot */
		if (malo_cmd_set_slot(sc, 1)) {
			device_printf(sc->sc_dev,
			    "setting short slot failed\n");
			return (ENXIO);
		}
	} else {
		/* set long slot */
		if (malo_cmd_set_slot(sc, 0)) {
			device_printf(sc->sc_dev, "setting long slot failed\n");
			return (ENXIO);
		}
	}

	return (0);
}

static void
malo_update_slot(struct ifnet *ifp)
{
	struct malo_softc *sc;
	struct ieee80211com *ic;

	sc = ifp->if_softc;
	ic = &sc->sc_ic;

	malo_set_slot(sc);

	if (ic->ic_opmode == IEEE80211_M_HOSTAP) {
		/* TODO */
	}
}

#ifdef MALO_DEBUG
static void
malo_hexdump(void *buf, int len)
{
	u_char b[16];
	int i, j, l;

	for (i = 0; i < len; i += l) {
		printf("%4i:", i);
		l = min(sizeof(b), len - i);
		bcopy((uint8_t *)buf + i, b, l);
		
		for (j = 0; j < sizeof(b); j++) {
			if (j % 2 == 0)
				printf(" ");
			if (j % 8 == 0)
				printf(" ");
			if (j < l)
				printf("%02x", (int)b[j]);
			else
				printf("  ");
		}
		printf("  |");
		for (j = 0; j < l; j++) {
			if (b[j] >= 0x20 && b[j] <= 0x7e)
				printf("%c", b[j]);
			else
				printf(".");
		}
		printf("|\n");
	}
}
#endif

static char *
malo_cmd_string(uint16_t cmd)
{
	int i;
	static char cmd_buf[16];
	static const struct {
		uint16_t	 cmd_code;
		char		*cmd_string;
	} cmds[] = {
		{ MALO_CMD_GET_HW_SPEC,		"GetHwSpecifications"	},
		{ MALO_CMD_SET_RADIO,		"SetRadio"		},
		{ MALO_CMD_SET_TXPOWER,		"SetTxPower"		},
		{ MALO_CMD_SET_ANTENNA,		"SetAntenna"		},
		{ MALO_CMD_SET_PRESCAN,		"SetPrescan"		},
		{ MALO_CMD_SET_POSTSCAN,	"SetPostscan"		},
		{ MALO_CMD_SET_CHANNEL,		"SetChannel"		},
		{ MALO_CMD_SET_RTS,		"SetRTS"		},
		{ MALO_CMD_SET_SLOT,		"SetSlot"		},
	};

	for (i = 0; i < sizeof(cmds) / sizeof(cmds[0]); i++)
		if ((le16toh(cmd) & 0x7fff) == cmds[i].cmd_code)
			return (cmds[i].cmd_string);

	snprintf(cmd_buf, sizeof(cmd_buf), "unknown %04x", cmd & 0x7fff);
	return (cmd_buf);
}

static char *
malo_cmd_string_result(uint16_t result)
{
	int i;
	static const struct {
		uint16_t	 result_code;
		char		*result_string;
	} results[] = {
		{ MALO_CMD_RESULT_OK,		"OK"		},
		{ MALO_CMD_RESULT_ERROR,	"general error"	},
		{ MALO_CMD_RESULT_NOSUPPORT,	"not supported" },
		{ MALO_CMD_RESULT_PENDING,	"pending"	},
		{ MALO_CMD_RESULT_BUSY,		"ignored"	},
		{ MALO_CMD_RESULT_PARTIALDATA,	"incomplete"	},
	};

	for (i = 0; i < sizeof(results) / sizeof(results[0]); i++)
		if (le16toh(result) == results[i].result_code)
			return (results[i].result_string);

	return ("unknown");
}

static int
malo_cmd_get_spec(struct malo_softc *sc)
{
	struct malo_cmdheader *hdr = sc->sc_cmd_mem;
	struct malo_hw_spec *spec;

	MALO_LOCK_ASSERT(sc);

	hdr->cmd = htole16(MALO_CMD_GET_HW_SPEC);
	hdr->size = htole16(sizeof(*hdr) + sizeof(*spec));
	hdr->seqnum = htole16(42);	/* the one and only */
	hdr->result = 0;
	spec = (struct malo_hw_spec *)(hdr + 1);

	bzero(spec, sizeof(*spec));
	memset(spec->PermanentAddress, 0xff, ETHER_ADDR_LEN);
	spec->CookiePtr = htole32(sc->sc_cookie_dmaaddr);

	bus_dmamap_sync(sc->sc_cmd_tag, sc->sc_cmd_map,
	    BUS_DMASYNC_PREWRITE | BUS_DMASYNC_PREREAD);

	if (malo_send_cmd_dma(sc, sc->sc_cmd_dmaaddr) != 0)
		return (ETIMEDOUT);

	/* XXX get the data from the buffer and feed it to ieee80211 */
	DPRINTF(sc, "get_hw_spec: V%x R%x, #WCB %d, #Mcast %d, Regcode %d, "
	    "#Ant %d\n", htole16(spec->HwVersion),
	    htole32(spec->FWReleaseNumber), htole16(spec->NumOfWCB),
	    htole16(spec->NumOfMCastAdr), htole16(spec->RegionCode),
	    htole16(spec->NumberOfAntenna));

	/* tell the DMA engine where our rings are */
	malo_mem_write4(sc, le32toh(spec->RxPdRdPtr) & 0xffff,
	    sc->sc_rxring.physaddr);
	malo_mem_write4(sc, le32toh(spec->RxPdWrPtr) & 0xffff,
	    sc->sc_rxring.physaddr);
	malo_mem_write4(sc, le32toh(spec->WcbBase0) & 0xffff,
	    sc->sc_txring.physaddr);

	/* save DMA RX pointers for later use */
	sc->sc_RxPdRdPtr = le32toh(spec->RxPdRdPtr) & 0xffff;
	sc->sc_RxPdWrPtr = le32toh(spec->RxPdWrPtr) & 0xffff;

	return (0);
}

static int
malo_cmd_set_wepkey(struct malo_softc *sc, struct ieee80211_key *wk,
    uint16_t wk_index)
{
	struct malo_cmdheader *hdr;
	struct malo_cmd_wepkey *body;

	MALO_LOCK_ASSERT(sc);

	hdr = sc->sc_cmd_mem;
	hdr->cmd = htole16(MALO_CMD_SET_WEPKEY);
	hdr->size = htole16(sizeof(*hdr) + sizeof(*body));
	hdr->seqnum = 1;
	hdr->result = 0;
	body = (struct malo_cmd_wepkey *)(hdr + 1);

	bzero(body, sizeof(*body));
	body->action = htole16(1);
	body->flags = 0;
	body->index = wk_index;
	body->len = wk->wk_keylen;
	memcpy(body->value, wk->wk_key, wk->wk_keylen);

	bus_dmamap_sync(sc->sc_cmd_tag, sc->sc_cmd_map,
	    BUS_DMASYNC_PREWRITE | BUS_DMASYNC_PREREAD);

	return (malo_send_cmd_dma(sc, sc->sc_cmd_dmaaddr));
}

static int
malo_cmd_set_prescan(struct malo_softc *sc)
{
	struct malo_cmdheader *hdr;

	MALO_LOCK_ASSERT(sc);

	hdr = sc->sc_cmd_mem;
	hdr->cmd = htole16(MALO_CMD_SET_PRESCAN);
	hdr->size = htole16(sizeof(*hdr));
	hdr->seqnum = 1;
	hdr->result = 0;

	bus_dmamap_sync(sc->sc_cmd_tag, sc->sc_cmd_map,
	    BUS_DMASYNC_PREWRITE | BUS_DMASYNC_PREREAD);

	return (malo_send_cmd_dma(sc, sc->sc_cmd_dmaaddr));
}

static int
malo_cmd_set_postscan(struct malo_softc *sc, uint8_t *macaddr, uint8_t ibsson)
{
	struct malo_cmdheader *hdr;
	struct malo_cmd_postscan *body;

	MALO_LOCK_ASSERT(sc);

	hdr = sc->sc_cmd_mem;
	hdr->cmd = htole16(MALO_CMD_SET_POSTSCAN);
	hdr->size = htole16(sizeof(*hdr) + sizeof(*body));
	hdr->seqnum = 1;
	hdr->result = 0;
	body = (struct malo_cmd_postscan *)(hdr + 1);

	bzero(body, sizeof(*body));
	memcpy(&body->bssid, macaddr, ETHER_ADDR_LEN);
	body->isibss = htole32(ibsson);

	bus_dmamap_sync(sc->sc_cmd_tag, sc->sc_cmd_map,
	    BUS_DMASYNC_PREWRITE | BUS_DMASYNC_PREREAD);

	return (malo_send_cmd_dma(sc, sc->sc_cmd_dmaaddr));
}

static int
malo_cmd_set_channel(struct malo_softc *sc, uint8_t channel)
{
	struct malo_cmdheader *hdr;
	struct malo_cmd_channel *body;

	MALO_LOCK_ASSERT(sc);

	hdr = sc->sc_cmd_mem;
	hdr->cmd = htole16(MALO_CMD_SET_CHANNEL);
	hdr->size = htole16(sizeof(*hdr) + sizeof(*body));
	hdr->seqnum = 1;
	hdr->result = 0;
	body = (struct malo_cmd_channel *)(hdr + 1);

	bzero(body, sizeof(*body));
	body->action = htole16(1);
	body->channel = channel;

	bus_dmamap_sync(sc->sc_cmd_tag, sc->sc_cmd_map,
	    BUS_DMASYNC_PREWRITE | BUS_DMASYNC_PREREAD);

	return (malo_send_cmd_dma(sc, sc->sc_cmd_dmaaddr));
}

static int
malo_cmd_set_antenna(struct malo_softc *sc, uint16_t antenna)
{
	struct malo_cmdheader *hdr;
	struct malo_cmd_antenna *body;

	MALO_LOCK_ASSERT(sc);

	hdr = sc->sc_cmd_mem;
	hdr->cmd = htole16(MALO_CMD_SET_ANTENNA);
	hdr->size = htole16(sizeof(*hdr) + sizeof(*body));
	hdr->seqnum = 1;
	hdr->result = 0;
	body = (struct malo_cmd_antenna *)(hdr + 1);

	bzero(body, sizeof(*body));
	body->action = htole16(antenna);
	if (antenna == 1)
		body->mode = htole16(0xffff);
	else
		body->mode = htole16(2);

	bus_dmamap_sync(sc->sc_cmd_tag, sc->sc_cmd_map,
	    BUS_DMASYNC_PREWRITE | BUS_DMASYNC_PREREAD);

	return (malo_send_cmd_dma(sc, sc->sc_cmd_dmaaddr));
}

static int
malo_cmd_set_radio(struct malo_softc *sc, uint16_t enable,
    uint16_t preamble_mode)
{
	struct malo_cmdheader *hdr;
	struct malo_cmd_radio *body;

	MALO_LOCK_ASSERT(sc);

	hdr = sc->sc_cmd_mem;
	hdr->cmd = htole16(MALO_CMD_SET_RADIO);
	hdr->size = htole16(sizeof(*hdr) + sizeof(*body));
	hdr->seqnum = 1;
	hdr->result = 0;
	body = (struct malo_cmd_radio *)(hdr + 1);

	bzero(body, sizeof(*body));
	body->action = htole16(1);
	body->preamble_mode = htole16(preamble_mode);
	body->enable = htole16(enable);

	bus_dmamap_sync(sc->sc_cmd_tag, sc->sc_cmd_map,
	    BUS_DMASYNC_PREWRITE | BUS_DMASYNC_PREREAD);

	return (malo_send_cmd_dma(sc, sc->sc_cmd_dmaaddr));
}

static int
malo_cmd_set_aid(struct malo_softc *sc, uint8_t *bssid, uint16_t associd)
{
	struct malo_cmdheader *hdr;
	struct malo_cmd_aid *body;

	MALO_LOCK_ASSERT(sc);

	hdr = sc->sc_cmd_mem;
	hdr->cmd = htole16(MALO_CMD_SET_AID);
	hdr->size = htole16(sizeof(*hdr) + sizeof(*body));
	hdr->seqnum = 1;
	hdr->result = 0;
	body = (struct malo_cmd_aid *)(hdr + 1);

	bzero(body, sizeof(*body));
	body->associd = htole16(associd);
	memcpy(&body->macaddr[0], bssid, IEEE80211_ADDR_LEN);

	bus_dmamap_sync(sc->sc_cmd_tag, sc->sc_cmd_map,
	    BUS_DMASYNC_PREWRITE | BUS_DMASYNC_PREREAD);

	return (malo_send_cmd_dma(sc, sc->sc_cmd_dmaaddr));
}

static int
malo_cmd_set_txpower(struct malo_softc *sc, unsigned int powerlevel)
{
	struct malo_cmdheader *hdr;
	struct malo_cmd_txpower *body;

	MALO_LOCK_ASSERT(sc);

	hdr = sc->sc_cmd_mem;
	hdr->cmd = htole16(MALO_CMD_SET_TXPOWER);
	hdr->size = htole16(sizeof(*hdr) + sizeof(*body));
	hdr->seqnum = 1;
	hdr->result = 0;
	body = (struct malo_cmd_txpower *)(hdr + 1);

	bzero(body, sizeof(*body));
	body->action = htole16(1);
	if (powerlevel >= 0 && powerlevel < 30)
		body->supportpowerlvl = htole16(5);	/* LOW */
	else if (powerlevel >= 30 && powerlevel < 60)
		body->supportpowerlvl = htole16(10);	/* MEDIUM */
	else
		body->supportpowerlvl = htole16(15);	/* HIGH */

	bus_dmamap_sync(sc->sc_cmd_tag, sc->sc_cmd_map,
	    BUS_DMASYNC_PREWRITE | BUS_DMASYNC_PREREAD);

	return (malo_send_cmd_dma(sc, sc->sc_cmd_dmaaddr));
}

static int
malo_cmd_set_rts(struct malo_softc *sc, uint32_t threshold)
{
	struct malo_cmdheader *hdr;

	MALO_LOCK_ASSERT(sc);

	hdr = sc->sc_cmd_mem;
	hdr->cmd = htole16(MALO_CMD_SET_RTS);
	hdr->size = htole16(sizeof(*hdr) + sizeof(threshold));
	hdr->seqnum = 1;
	hdr->result = 0;

	*(uint32_t *)(hdr + 1) = htole32(threshold);

	bus_dmamap_sync(sc->sc_cmd_tag, sc->sc_cmd_map,
	    BUS_DMASYNC_PREWRITE | BUS_DMASYNC_PREREAD);

	return (malo_send_cmd_dma(sc, sc->sc_cmd_dmaaddr));
}

static int
malo_cmd_set_slot(struct malo_softc *sc, uint8_t slot)
{
	struct malo_cmdheader *hdr;
	struct malo_cmd_slot *body;

	MALO_LOCK_ASSERT(sc);

	hdr = sc->sc_cmd_mem;
	hdr->cmd = htole16(MALO_CMD_SET_SLOT);
	hdr->size = htole16(sizeof(*hdr) + sizeof(*body));
	hdr->seqnum = 1;
	hdr->result = 0;
	body = (struct malo_cmd_slot *)(hdr + 1);

	bzero(body, sizeof(*body));
	body->action = htole16(1);
	body->slot = slot;

	bus_dmamap_sync(sc->sc_cmd_tag, sc->sc_cmd_map,
	    BUS_DMASYNC_PREWRITE | BUS_DMASYNC_PREREAD);

	return (malo_send_cmd_dma(sc, sc->sc_cmd_dmaaddr));
}

static int
malo_cmd_set_rate(struct malo_softc *sc, uint8_t rate)
{
	struct ieee80211com *ic;
	struct malo_cmdheader *hdr;
	struct malo_cmd_rate *body;
	int i;

	MALO_LOCK_ASSERT(sc);

	ic = &sc->sc_ic;
	hdr = sc->sc_cmd_mem;
	hdr->cmd = htole16(MALO_CMD_SET_RATE);
	hdr->size = htole16(sizeof(*hdr) + sizeof(*body));
	hdr->seqnum = 1;
	hdr->result = 0;
	body = (struct malo_cmd_rate *)(hdr + 1);

	bzero(body, sizeof(*body));

	if (ic->ic_opmode == IEEE80211_M_HOSTAP) {
		/* TODO */
	} else {
		body->aprates[0] = 2;
		body->aprates[1] = 4;
		body->aprates[2] = 11;
		body->aprates[3] = 22;
		if (ic->ic_curmode == IEEE80211_MODE_11G) {
			body->aprates[4] = 0;
			body->aprates[5] = 12;
			body->aprates[6] = 18;
			body->aprates[7] = 24;
			body->aprates[8] = 36;
			body->aprates[9] = 48;
			body->aprates[10] = 72;
			body->aprates[11] = 96;
			body->aprates[12] = 108;
		}
	}

	if (rate != 0) {
		/* fixed rate */
		for (i = 0; i < 13; i++) {
			if (body->aprates[i] == rate) {
				body->rateindex = i;
				body->dataratetype = 1;
				break;
			}
		}
	}

	bus_dmamap_sync(sc->sc_cmd_tag, sc->sc_cmd_map,
	    BUS_DMASYNC_PREWRITE | BUS_DMASYNC_PREREAD);

	return (malo_send_cmd_dma(sc, sc->sc_cmd_dmaaddr));
}

static device_method_t malo_methods[] = {
	/* Device interface */
	DEVMETHOD(device_probe,		malo_probe),
	DEVMETHOD(device_attach,	malo_attach),
	DEVMETHOD(device_detach,	malo_detach),
	DEVMETHOD(device_suspend,	malo_pci_suspend),
	DEVMETHOD(device_resume,	malo_pci_resume),
	DEVMETHOD(device_shutdown,	malo_pci_shutdown),

	{ NULL, NULL }
};

static driver_t malo_driver = {
	"malo",
	malo_methods,
	sizeof(struct malo_softc)
};

static devclass_t malo_devclass;
DRIVER_MODULE(malo, pci, malo_driver, malo_devclass, 0, 0);
DRIVER_MODULE(malo, cardbus, malo_driver, malo_devclass, 0, 0);
MODULE_DEPEND(malo, pci, 1, 1, 1);
MODULE_DEPEND(malo, wlan, 1, 1, 1);
